# pyright: reportInvalidTypeForm=false
# pylint: disable=too-many-lines
"""Pydantic classes for the data model."""

from datetime import datetime
from enum import Enum
from typing import (
    Annotated,
    Any,
    Callable,
    Collection,
    Iterator,
    Literal,
    Optional,
    Self,
    TypeVar,
    Union,
)
from uuid import UUID

import edtf  # type: ignore
from pydantic import (
    AnyUrl,
    ConfigDict,
    Field,
    StringConstraints,
    field_validator,
    model_validator,
)
from pydantic import (
    BaseModel as PydanticBaseModel,
)
from pydantic.main import IncEx

SiblingModel = TypeVar("SiblingModel", bound=PydanticBaseModel)
FieldValue = Union[str, int, "BaseModel"]
T = TypeVar("T")


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(extra="forbid", frozen=True)

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = True,  # Overriding this
        exclude_unset: bool = True,  # Overriding this
        exclude_defaults: bool = True,  # Overriding this
        exclude_none: bool = True,  # Overriding this
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        return super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )

    def model_dump_json(
        self,
        *,
        indent: int | None = None,
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = True,  # Overriding this
        exclude_unset: bool = True,  # Overriding this
        exclude_defaults: bool = True,  # Overriding this
        exclude_none: bool = True,  # Overriding this
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> str:
        return super().model_dump_json(
            indent=indent,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )

    # type: ignore
    def convert(self, cls: type[SiblingModel], **updated_data: Any) -> SiblingModel:
        data = self.model_dump() | updated_data
        return cls(**data)

    def deep_get(
        self, *names: str, cls: type[T], exclude: Collection[str] = tuple()
    ) -> Iterator[T]:
        for field_name in self.model_fields_set:
            if field_name in exclude:
                continue

            field_value: tuple[FieldValue] | FieldValue = getattr(self, field_name)
            field_value_tuple: tuple[FieldValue] = (
                field_value if isinstance(field_value, tuple) else (field_value,)
            )

            for obj in field_value_tuple:
                if isinstance(obj, cls) and (len(names) == 0 or field_name in names):
                    yield obj

                if isinstance(obj, BaseModel) and field_name not in exclude:
                    yield from obj.deep_get(*names, cls=cls, exclude=exclude)


NonEmptyStr = Annotated[
    str,
    StringConstraints(min_length=1),
]

Ark = Annotated[
    str,
    StringConstraints(pattern=r"^ark:/21198/([a-z]|[0-9])+"),
]


class ControlledTerm(BaseModel):
    id: NonEmptyStr
    label: NonEmptyStr


# generated by datamodel-codegen:
#   filename:  agent.schema.json
#   timestamp: 2025-05-28T17:21:15+00:00


class Iso(BaseModel):
    not_before: str
    not_after: str | None = None

    @field_validator("not_before", mode="after")
    @classmethod
    def is_date_string(cls, value: str) -> str:
        assert isinstance(edtf.parse_edtf(value), edtf.Date)
        return value

    @field_validator("not_after", mode="after")
    @classmethod
    def is_date_string_or_none(cls, value: str) -> str:
        assert value is None or isinstance(edtf.parse_edtf(value), edtf.Date)
        return value

    def years(self) -> range:
        start = int(edtf.parse_edtf(self.not_before).year)
        end = int(edtf.parse_edtf(self.not_after or self.not_before).year)
        return range(start, end + 1)


class Date(BaseModel):
    value: Optional[NonEmptyStr] = None
    iso: Iso


class RelatedConceptSource(Enum):
    HAF = "HAF"
    LoC = "LoC"
    Perseus = "Perseus"
    Pinakes = "Pinakes"
    Pleiades = "Pleiades"
    Syriaca = "Syriaca"
    VIAF = "VIAF"


class RelConItem(BaseModel):
    label: NonEmptyStr
    uri: AnyUrl
    source: RelatedConceptSource


class ReferenceNumberSource(Enum):
    CPG = "CPG"
    CPL = "CPL"
    TLG = "TLG"


class RefnoItem(BaseModel):
    label: NonEmptyStr
    idno: NonEmptyStr
    source: ReferenceNumberSource


class BibItem(BaseModel):
    id: UUID
    type: ControlledTerm
    shortcode: Optional[str] = (
        None  # TODO: required in JSON schema but not found in all data
    )
    citation: Optional[str] = (
        None  # TODO: required in JSON schema but not found in all data
    )
    range: Optional[NonEmptyStr] = None
    alt_shelf: Optional[NonEmptyStr] = None
    url: Optional[NonEmptyStr] = None
    note: tuple[str, ...] = tuple()


class RelItem(BaseModel):
    id: NonEmptyStr
    label: NonEmptyStr


class RelAgentItem(BaseModel):
    id: Ark = Field(..., description="A unique Archival Resource Key (ARK)")
    source: tuple[str, ...] = tuple()
    rel: tuple[RelItem, ...]


class RelPlaceItem(BaseModel):
    id: Ark = Field(..., description="A unique Archival Resource Key (ARK)")
    source: tuple[str, ...] = tuple()
    rel: tuple[RelItem, ...]


class CataloguerItem(BaseModel):
    message: NonEmptyStr
    contributor: Optional[NonEmptyStr] = Field(
        None,
        description="The person responsible for the intellectual contributions made to the record via this change",
    )
    added_by: NonEmptyStr = Field(
        ...,
        description="The user account responsible for editing the JSON record, likely supplied by the content management system",
    )
    timestamp: datetime


class Agent(BaseModel):
    ark: Ark
    type: ControlledTerm
    pref_name: NonEmptyStr
    alt_name: tuple[NonEmptyStr, ...] = tuple()
    desc: Optional[NonEmptyStr] = None
    gender: Optional[ControlledTerm] = None
    birth: Optional[Date] = None
    death: Optional[Date] = None
    floruit: Optional[Date] = None
    rel_con: tuple[RelConItem, ...] = tuple()
    refno: tuple[RefnoItem, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()
    note: tuple[str, ...] = tuple()
    rel_agent: tuple[RelAgentItem, ...] = tuple()
    rel_place: tuple[RelPlaceItem, ...] = tuple()
    internal: tuple[str, ...] = tuple()
    cataloguer: tuple[CataloguerItem, ...] = tuple()


# generated by datamodel-codegen:
#   filename:  work.schema.json
#   timestamp: 2025-05-28T17:23:18+00:00


class Creation(BaseModel):
    value: Optional[NonEmptyStr] = None
    iso: Iso


# TODO: not attested
class Incipit(BaseModel):
    value: NonEmptyStr
    translation: tuple[str, ...] = tuple()
    source: tuple[str, ...] = tuple()


# TODO not attested
class Explicit(BaseModel):
    value: NonEmptyStr
    translation: tuple[str, ...] = tuple()
    source: tuple[str, ...] = tuple()


# TODO not attested
class RelWorkItem(BaseModel):
    id: Ark
    source: tuple[str, ...] = tuple()
    rel: tuple[RelItem, ...]


class AssocNameItem(BaseModel):
    id: Optional[Ark] = None
    agent_record: Optional[Agent] = None
    value: Optional[NonEmptyStr] = None
    as_written: Optional[NonEmptyStr] = None
    role: ControlledTerm
    note: tuple[str, ...] = tuple()


class AssocNameItemMerged(AssocNameItem):
    agent_record: Optional[Agent] = None

    @model_validator(mode="after")
    def agent_record_loaded(self) -> Self:
        if self.id and not self.agent_record:
            raise ValueError("has agent id but agent_record not loaded")
        return self


class AssocNameItemUnmerged(AssocNameItem):
    agent_record: None = None


class ConceptualWork(BaseModel):
    ark: Ark
    pref_title: NonEmptyStr
    orig_lang: Optional[ControlledTerm] = None
    orig_lang_title: Optional[NonEmptyStr] = None
    alt_title: tuple[NonEmptyStr, ...] = tuple()
    desc: Optional[NonEmptyStr] = None
    genre: tuple[ControlledTerm, ...] = tuple()
    creator: tuple[AssocNameItemUnmerged, ...] | tuple[AssocNameItemMerged, ...] = (
        tuple()
    )
    creation: Optional[Creation] = None
    incipit: Optional[Incipit] = None
    explicit: Optional[Explicit] = None
    rel_con: tuple[RelConItem, ...] = tuple()
    refno: tuple[RefnoItem, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()
    note: tuple[str, ...] = tuple()
    rel_work: tuple[RelWorkItem, ...] = tuple()
    rel_agent: tuple[RelAgentItem, ...] = tuple()
    rel_place: tuple[RelPlaceItem, ...] = tuple()
    internal: tuple[str, ...] = tuple()
    cataloguer: tuple[CataloguerItem, ...] = tuple()


class ConceptualWorkUnmerged(ConceptualWork):
    creator: tuple[AssocNameItemUnmerged, ...] = tuple()


class ConceptualWorkMerged(ConceptualWork):
    creator: tuple[AssocNameItemMerged, ...] = tuple()


# generated by datamodel-codegen:
#   filename:  text_unit.schema.json
#   timestamp: 2025-05-28T17:22:58+00:00


class WorkBriefCreator(BaseModel):
    id: Ark
    agent_record: Agent


class WorkStub(BaseModel):
    id: Ark


class WorkBrief(BaseModel):
    desc_title: Optional[NonEmptyStr] = None
    creator: tuple[Ark, ...] | tuple[WorkBriefCreator, ...] = tuple()
    genre: tuple[ControlledTerm, ...] = tuple()


class WorkBriefUnmerged(WorkBrief):
    creator: tuple[Ark, ...] = tuple()


class WorkBriefMerged(WorkBrief):
    creator: tuple[WorkBriefCreator, ...] = tuple()


class ExcerptItem(BaseModel):
    type: ControlledTerm
    locus: Optional[NonEmptyStr] = None
    as_written: Optional[NonEmptyStr] = None
    translation: tuple[str, ...] = tuple()
    note: tuple[str, ...] = tuple()


class Contents(BaseModel):
    label: Optional[NonEmptyStr] = None
    work_id: Optional[Ark] = None
    locus: Optional[NonEmptyStr] = None
    note: tuple[str, ...] = tuple()

    @model_validator(mode="after")
    def has_label_or_work_id(self) -> Self:
        if not (self.label or self.work_id):
            raise ValueError("`Contents.label` and `.work_id` cannot both be None")
        return self


class ContentsUnmerged(Contents):
    pass


class ContentsMerged(Contents):
    pref_title: Optional[NonEmptyStr] = None


class WorkWitItem(BaseModel):
    work: WorkBrief | WorkStub | ConceptualWorkMerged
    alt_title: Optional[str] = Field(
        None,
        description="An alternative title providing witness-specific information about the work",
    )
    as_written: Optional[NonEmptyStr] = None
    locus: Optional[NonEmptyStr] = None
    excerpt: tuple[ExcerptItem, ...] = tuple()
    contents: tuple[ContentsUnmerged, ...] | tuple[ContentsMerged, ...] = tuple()
    note: tuple[str, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()


class WorkWitItemUnmerged(WorkWitItem):
    work: WorkBriefUnmerged | WorkStub
    contents: tuple[ContentsUnmerged, ...] = tuple()


class WorkWitItemMerged(WorkWitItem):
    work: WorkBriefMerged | ConceptualWorkMerged
    contents: tuple[ContentsMerged, ...] = tuple()


class ScriptItem(BaseModel):
    id: NonEmptyStr
    label: NonEmptyStr
    writing_system: str


class Place(BaseModel):
    ark: Ark
    pref_name: NonEmptyStr
    alt_name: tuple[NonEmptyStr, ...] = tuple()


class AssocPlaceItem(BaseModel):
    id: Optional[Ark] = None
    value: Optional[NonEmptyStr] = None
    as_written: Optional[NonEmptyStr] = None
    event: ControlledTerm
    note: tuple[str, ...] = tuple()


class AssocPlaceItemUnmerged(AssocPlaceItem):
    pass


class AssocPlaceItemMerged(AssocPlaceItem):
    place_record: Place | None = None

    @model_validator(mode="after")
    def has_place_record_if_possible(self) -> Self:
        if self.id and not self.place_record:
            raise ValueError(f"{self} has id {self.id} but no place_record loaded")
        return self


class AssocDateItem(BaseModel):
    value: NonEmptyStr
    iso: Optional[Iso] = None
    type: ControlledTerm
    as_written: Optional[NonEmptyStr] = None
    note: tuple[str, ...] = tuple()


class ParaItem(BaseModel):
    type: ControlledTerm
    subtype: tuple[ControlledTerm, ...] = tuple()
    locus: NonEmptyStr
    lang: tuple[ControlledTerm, ...] = Field(..., min_length=1)
    script: tuple[ScriptItem, ...] = tuple()
    label: Optional[NonEmptyStr] = None
    as_written: Optional[NonEmptyStr] = None
    translation: tuple[str, ...] = tuple()
    assoc_name: tuple[AssocNameItemUnmerged, ...] | tuple[AssocNameItemMerged, ...] = (
        tuple()
    )
    assoc_place: (
        tuple[AssocPlaceItemUnmerged, ...] | tuple[AssocPlaceItemMerged, ...]
    ) = tuple()
    assoc_date: tuple[AssocDateItem, ...] = tuple()
    note: tuple[str, ...] = tuple()


class ParaItemUnmerged(ParaItem):
    assoc_name: tuple[AssocNameItemUnmerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemUnmerged, ...] = tuple()


class ParaItemMerged(ParaItem):
    assoc_name: tuple[AssocNameItemMerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemMerged, ...] = tuple()


class NoteItem(BaseModel):
    type: ControlledTerm
    value: Optional[NonEmptyStr]


class DescProvenanceProgramItem(BaseModel):
    label: str
    description: str


class DescProvenance(BaseModel):
    program: tuple[DescProvenanceProgramItem, ...] = tuple()
    rights: Optional[str] = None


class TextUnit(BaseModel):
    ark: Ark
    reconstruction: bool
    label: NonEmptyStr
    summary: Optional[NonEmptyStr] = None
    locus: Optional[NonEmptyStr] = None
    lang: tuple[ControlledTerm, ...] = Field(..., min_length=1)
    work_wit: tuple[WorkWitItemUnmerged, ...] | tuple[WorkWitItemMerged, ...] = Field(
        ..., min_length=1
    )
    para: tuple[ParaItemUnmerged, ...] | tuple[ParaItemMerged, ...] = tuple()
    features: tuple[ControlledTerm, ...] = tuple()
    note: tuple[NoteItem, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()
    cataloguer: tuple[CataloguerItem, ...] = tuple()
    reconstructed_from: tuple[Ark, ...] = tuple()
    parent: tuple[Ark, ...] = Field(..., min_length=1)
    internal: tuple[str, ...] = tuple()
    desc_provenance: Optional[DescProvenance] = None


class TextUnitUnmerged(TextUnit):
    work_wit: tuple[WorkWitItemUnmerged, ...] = Field(..., min_length=1)
    para: tuple[ParaItemUnmerged, ...] = tuple()


class TextUnitMerged(TextUnit):
    work_wit: tuple[WorkWitItemMerged, ...] = Field(..., min_length=1)
    para: tuple[ParaItemMerged, ...] = tuple()


# generated by datamodel-codegen:
#   filename:  layer.schema.json
#   timestamp: 2025-05-28T17:21:54+00:00

# from __future__ import annotations

# from datetime import datetime
# from typing import List, Optional
# from uuid import UUID

# from pydantic import AnyUrl, BaseModel, Field, constr


class WritingItem(BaseModel):
    script: tuple[ScriptItem, ...] = Field(..., min_length=1)
    locus: Optional[NonEmptyStr] = None
    note: tuple[str, ...] = tuple()


class InkItem(BaseModel):
    locus: Optional[NonEmptyStr] = None
    color: tuple[str, ...] = tuple()
    note: tuple[str, ...] = tuple()


class LayoutItem(BaseModel):
    locus: Optional[NonEmptyStr] = None
    columns: Optional[str] = None
    lines: Optional[str] = None
    dim: Optional[NonEmptyStr] = Field(
        None,
        description="A string expression of an object's dimensions, whether manuscript block, folio, or writing area",
    )
    note: tuple[str, ...] = tuple()


class LayerTextUnit(BaseModel):
    id: Ark = Field(..., description="A unique Archival Resource Key (ARK)")
    label: NonEmptyStr
    locus: Optional[NonEmptyStr] = None


class LayerTextUnitUnmerged(LayerTextUnit):
    pass


class LayerTextUnitMerged(LayerTextUnit):
    text_unit_record: TextUnitMerged


class Ms(BaseModel):
    label: NonEmptyStr
    id: Optional[Ark] = None
    url: Optional[AnyUrl] = None


class RelatedMs(BaseModel):
    type: ControlledTerm
    label: NonEmptyStr
    note: tuple[str, ...] = tuple()
    mss: tuple[Ms, ...] = Field(..., min_length=1)


class InscribedLayer(BaseModel):
    ark: Ark
    reconstruction: bool
    state: ControlledTerm = Field(
        ..., description="The current state of the inscribed layer"
    )
    label: NonEmptyStr
    locus: Optional[NonEmptyStr] = None
    summary: Optional[NonEmptyStr] = None
    extent: Optional[NonEmptyStr] = Field(
        None,
        description="The extent, expressed in number of folios, which comprise the manuscript object",
    )
    writing: tuple[WritingItem, ...] = Field(..., min_length=1)
    ink: tuple[InkItem, ...] = tuple()
    layout: tuple[LayoutItem, ...] = tuple()
    text_unit: tuple[LayerTextUnitUnmerged, ...] | tuple[LayerTextUnitMerged, ...] = (
        Field(..., min_length=1)
    )
    para: tuple[ParaItemUnmerged, ...] | tuple[ParaItemMerged, ...] = tuple()
    assoc_date: tuple[AssocDateItem, ...] = tuple()
    assoc_name: tuple[AssocNameItemUnmerged, ...] | tuple[AssocNameItemMerged, ...] = (
        tuple()
    )
    assoc_place: (
        tuple[AssocPlaceItemUnmerged, ...] | tuple[AssocPlaceItemMerged, ...]
    ) = tuple()
    features: tuple[ControlledTerm, ...] = tuple()
    related_mss: tuple[RelatedMs, ...] = tuple()
    note: tuple[NoteItem, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()
    cataloguer: tuple[CataloguerItem, ...] = tuple()
    reconstructed_from: tuple[Ark, ...] = tuple()
    parent: tuple[Ark, ...] = Field(..., min_length=1)
    internal: tuple[str, ...] = tuple()
    desc_provenance: Optional[DescProvenance] = None


class InscribedLayerUnmerged(InscribedLayer):
    text_unit: tuple[LayerTextUnitUnmerged, ...] = Field(..., min_length=1)
    para: tuple[ParaItemUnmerged, ...] = tuple()
    assoc_name: tuple[AssocNameItemUnmerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemUnmerged, ...] = tuple()


class InscribedLayerMerged(InscribedLayer):
    text_unit: tuple[LayerTextUnitMerged, ...] = Field(..., min_length=1)
    para: tuple[ParaItemMerged, ...] = tuple()
    assoc_name: tuple[AssocNameItemMerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemMerged, ...] = tuple()


# generated by datamodel-codegen:
#   filename:  ms_obj.schema.json
#   timestamp: 2025-05-28T17:22:16+00:00

# from __future__ import annotations

# from datetime import datetime
# from enum import Enum
# from typing import List, Optional
# from uuid import UUID

# from pydantic import AnyUrl, BaseModel, Field, constr


class ManuscriptLayer(BaseModel):
    id: Ark
    label: NonEmptyStr
    type: ControlledTerm
    layer_record: Optional[InscribedLayerMerged] = None
    locus: Optional[NonEmptyStr] = None


class ManuscriptLayerUnmerged(ManuscriptLayer):
    layer_record: None = None


class ManuscriptLayerMerged(ManuscriptLayer):
    layer_record: InscribedLayerMerged


class UndertextManuscriptLayerMerged(ManuscriptLayer):
    id: Ark = Field(..., serialization_alias="uto_layer_ark")

    layer_record: None = None

    uto_ms_ark: Optional[Ark] = None

    script: tuple[str, ...] = tuple()
    lang: tuple[str, ...] = tuple()
    orig_date: tuple[AssocDateItem, ...] = tuple()

    @model_validator(mode="after")
    def undertext_type(self) -> Self:
        if self.type != ControlledTerm(id="undertext", label="Undertext"):
            raise ValueError(
                'UndertextManuscriptLayerMerged records must have type="undertext"'
            )
        return self


class Part(BaseModel):
    label: Optional[NonEmptyStr] = None
    summary: Optional[NonEmptyStr] = None
    locus: Optional[NonEmptyStr] = None
    support: tuple[ControlledTerm, ...] = Field(..., min_length=1)
    extent: Optional[NonEmptyStr] = Field(
        None,
        description="The extent, expressed in number of folios, which comprise the manuscript object",
    )
    dim: Optional[NonEmptyStr] = Field(
        None,
        description="A string expression of an object's dimensions, whether manuscript block, folio, or writing area",
    )
    layer: tuple[ManuscriptLayerUnmerged, ...] = Field(..., min_length=1)
    para: tuple[ParaItemUnmerged, ...] | tuple[ParaItemMerged, ...] = tuple()
    note: tuple[NoteItem, ...] = tuple()
    related_mss: tuple[RelatedMs, ...] = tuple()


class PartUnmerged(Part):
    para: tuple[ParaItemUnmerged, ...] = tuple()


class PartMerged(Part):
    layer: tuple[ManuscriptLayerUnmerged, ...] = Field(..., max_length=0, exclude=True)
    ot_layer: tuple[ManuscriptLayerMerged, ...] = Field(..., min_length=1)
    guest_layer: tuple[ManuscriptLayerMerged, ...] = tuple()
    uto: tuple[UndertextManuscriptLayerMerged, ...] = tuple()
    para: tuple[ParaItemMerged, ...] = tuple()


class LocationItem(BaseModel):
    id: NonEmptyStr
    collection: Optional[str] = None
    repository: NonEmptyStr


class ViscodexItem(BaseModel):
    type: ControlledTerm
    label: NonEmptyStr
    url: AnyUrl


class TextDirection(Enum):
    right_to_left = "right-to-left"
    top_to_bottom = "top-to-bottom"
    bottom_to_top = "bottom-to-top"


class Behavior(Enum):
    paged = "paged"
    individuals = "individuals"


class IiifItem(BaseModel):
    type: ControlledTerm
    manifest: AnyUrl
    text_direction: Optional[TextDirection] = None
    behavior: Optional[Behavior] = None
    thumbnail: Optional[AnyUrl] = None

    @field_validator("manifest", mode="before")
    @classmethod
    def remove_ingest_from_url(cls, value: AnyUrl) -> AnyUrl:
        return AnyUrl(
            str(value).replace("ingest.iiif.library.ucla.edu", "iiif.library.ucla.edu")
        )


# TODO only used in unattested class DigProvenance
class Image(BaseModel):
    camera_operator: tuple[str, ...] = tuple()
    imaging_date: Optional[str] = None
    delivery: Optional[str] = None
    msi_processing: tuple[str, ...] = Field(
        tuple(),
        description="Names of the person(s) responsible for processing the MSI data for this record",
    )
    program: Optional[str] = None
    condition_category: Optional[str] = None
    note: tuple[str, ...] = tuple()


# TODO only used in unattested class DigProvenance
class Description(BaseModel):
    program: str
    note: tuple[str, ...] = tuple()


# TODO class not attested in data, replace by DescProvenance and ImageProvenance?
class DigProvenance(BaseModel):
    images: tuple[Image, ...] = tuple()
    descriptions: tuple[Description, ...] = tuple()
    image_rights: NonEmptyStr
    metadata_rights: NonEmptyStr


# TODO attested in all data but only found in the export_test schema
class ImageProvenanceProgramItem(BaseModel):
    label: Optional[str] = None
    description: Optional[str] = None
    camera_operator: tuple[str, ...] = tuple()
    imaging_date: Optional[str] = None
    delivery: Optional[str] = None
    msi_processing: tuple[str, ...] = Field(
        tuple(),
        description="Names of the person(s) responsible for processing the MSI data for this record",
    )
    condition_category: Optional[str] = None
    note: tuple[str, ...] = tuple()

    # TODO: not found in schemas, attested in data
    imaging_system: Optional[str] = None


class ImageProvenance(BaseModel):
    program: tuple[ImageProvenanceProgramItem, ...] = tuple()
    rights: Optional[NonEmptyStr] = (
        None  # TODO schema defines as not optional but gives None as default
    )


class ReconstructedFrom(BaseModel):
    id: Ark
    shelfmark: NonEmptyStr


class ManuscriptObject(BaseModel):
    ark: Ark = Field(..., description="A unique Archival Resource Key (ARK)")
    reconstruction: bool
    type: ControlledTerm
    shelfmark: NonEmptyStr
    summary: Optional[NonEmptyStr] = None
    extent: Optional[NonEmptyStr] = Field(
        None,
        description="The extent, expressed in number of folios, which comprise the manuscript object",
    )
    weight: Optional[NonEmptyStr] = Field(
        None, description="A string expression of the manuscript object's weight"
    )
    dim: Optional[NonEmptyStr] = Field(
        None,
        description="A string expression of an object's dimensions, whether manuscript block, folio, or writing area",
    )
    state: ControlledTerm = Field(
        ..., description="The current state of the manuscript object"
    )
    fol: Optional[NonEmptyStr] = Field(
        None,
        description="A string expressing the foliation of the object in a semi-controlled format",
    )
    coll: Optional[NonEmptyStr] = Field(
        None,
        description="A string expressing the collation of the object following a semi-controlled format, such as a collational formula",
    )
    features: tuple[ControlledTerm, ...] = tuple()
    part: tuple[Part, ...] | tuple[PartUnmerged, ...] | tuple[PartMerged, ...] = Field(
        ..., min_length=1
    )
    layer: tuple[ManuscriptLayerUnmerged, ...] = tuple()
    para: tuple[ParaItemUnmerged, ...] | tuple[ParaItemMerged, ...] = tuple()
    location: tuple[LocationItem, ...]
    assoc_date: tuple[AssocDateItem, ...] = tuple()
    assoc_name: tuple[AssocNameItemMerged, ...] | tuple[AssocNameItemUnmerged, ...] = (
        tuple()
    )
    assoc_place: (
        tuple[AssocPlaceItemUnmerged, ...] | tuple[AssocPlaceItemMerged, ...]
    ) = tuple()
    note: tuple[NoteItem, ...] = tuple()
    related_mss: tuple[RelatedMs, ...] = tuple()
    viscodex: tuple[ViscodexItem, ...] = tuple()
    bib: tuple[BibItem, ...] = tuple()
    iiif: tuple[IiifItem, ...] = tuple()
    internal: tuple[str, ...] = tuple()
    cataloguer: tuple[CataloguerItem, ...] = tuple()
    reconstructed_from: tuple[Ark, ...] | tuple[ReconstructedFrom, ...] = tuple()

    # in regular schemas but not export_test schema, unattested in data
    # think this got replaced with desc_provenance and/or image_provenance
    dig_provenance: Optional[DigProvenance] = None

    # found in export_test schema but not other schemas, seems to be used in all data
    desc_provenance: Optional[DescProvenance] = None
    image_provenance: Optional[ImageProvenance] = None

    # not in the schemas, found in the export_test data but not elsewhere
    has_bind: Optional[bool] = None


class ManuscriptObjectUnmerged(ManuscriptObject):
    part: tuple[PartUnmerged, ...]
    assoc_name: tuple[AssocNameItemUnmerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemUnmerged, ...] = tuple()
    para: tuple[ParaItemUnmerged, ...] = tuple()
    reconstructed_from: tuple[Ark, ...] = tuple()


class ManuscriptObjectMerged(ManuscriptObject):
    part: tuple[PartMerged, ...]

    layer: tuple[ManuscriptLayerUnmerged, ...] = Field(
        tuple(), max_length=0, exclude=True
    )
    ot_layer: tuple[ManuscriptLayerMerged, ...] = tuple()
    guest_layer: tuple[ManuscriptLayerMerged, ...] = tuple()
    uto: tuple[UndertextManuscriptLayerMerged, ...] = tuple()

    assoc_name: tuple[AssocNameItemMerged, ...] = tuple()
    assoc_place: tuple[AssocPlaceItemMerged, ...] = tuple()

    para: tuple[ParaItemMerged, ...] = tuple()
    reconstructed_from: tuple[ReconstructedFrom, ...] = tuple()



# pylint: disable=all
# mypy: disallow_untyped_defs=False


class MockResponse:
    def __init__(self, status_code, json_data):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


GOOD_MANIFEST = MockResponse(
    200,
    {
        "@context": "http://iiif.io/api/presentation/2/context.json",
        "label": "Sinai Arabic 352. Mimars and Lives of Saints : manuscript, 1200. St. Catherine's Monastery, Sinai, Egypt",
        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest",
        "@type": "sc:Manifest",
        "sequences": [
            {
                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/sequence/normal",
                "@type": "sc:Sequence",
                "canvases": [
                    {
                        "@type": "sc:Canvas",
                        "label": "Front Board Outside",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                        "width": 5332,
                        "height": 7006,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/hm957748",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fhm957748/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fhm957748",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                    {
                        "@type": "sc:Canvas",
                        "label": "f. 001r",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/zw07hs0c",
                        "width": 5332,
                        "height": 7008,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/zw07hs0c",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/zw07hs0c",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fzw07hs0c/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fzw07hs0c",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                    {
                        "@type": "sc:Canvas",
                        "label": "f. 049v",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/dw84c98r",
                        "width": 5332,
                        "height": 7008,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/dw84c98r",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/dw84c98r",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fdw84c98r/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fdw84c98r",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                ],
            }
        ],
    },
)

MANIFEST_WITHOUT_F001R = MockResponse(
    200,
    {
        "@context": "http://iiif.io/api/presentation/2/context.json",
        "label": "Sinai Arabic 352. Mimars and Lives of Saints : manuscript, 1200. St. Catherine's Monastery, Sinai, Egypt",
        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest",
        "@type": "sc:Manifest",
        "sequences": [
            {
                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/sequence/normal",
                "@type": "sc:Sequence",
                "canvases": [
                    {
                        "@type": "sc:Canvas",
                        "label": "Front Board Outside",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                        "width": 5332,
                        "height": 7006,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/hm957748",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fhm957748/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fhm957748",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                    {
                        "@type": "sc:Canvas",
                        "label": "f. 001v",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/zw07hs0c",
                        "width": 5332,
                        "height": 7008,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/zw07hs0c",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/zw07hs0c",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fzw07hs0c/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fzw07hs0c",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                    {
                        "@type": "sc:Canvas",
                        "label": "f. 049v",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/dw84c98r",
                        "width": 5332,
                        "height": 7008,
                        "images": [
                            {
                                "@type": "oa:Annotation",
                                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/annotation/dw84c98r",
                                "motivation": "sc:painting",
                                "on": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/dw84c98r",
                                "resource": {
                                    "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fdw84c98r/full/600,/0/default.jpg",
                                    "@type": "dctypes:Image",
                                    "format": "image/jpeg",
                                    "service": {
                                        "@context": "http://iiif.io/api/image/2/context.json",
                                        "@id": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fdw84c98r",
                                        "profile": "http://iiif.io/api/image/2/level0.json",
                                    },
                                },
                            }
                        ],
                    },
                ],
            }
        ],
    },
)

MANIFEST_WITHOUT_IMAGES = MockResponse(
    200,
    {
        "@context": "http://iiif.io/api/presentation/2/context.json",
        "label": "Sinai Arabic 352. Mimars and Lives of Saints : manuscript, 1200. St. Catherine's Monastery, Sinai, Egypt",
        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest",
        "@type": "sc:Manifest",
        "sequences": [
            {
                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/sequence/normal",
                "@type": "sc:Sequence",
                "canvases": [
                    {
                        "@type": "sc:Canvas",
                        "label": "Front Board Outside",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                        "width": 5332,
                        "height": 7006,
                        "images": [],
                    },
                ],
            }
        ],
    },
)

BAD_MANIFEST = MockResponse(
    200,
    {
        "@context": "http://iiif.io/api/presentation/2/context.json",
        "label": "Sinai Arabic 352. Mimars and Lives of Saints : manuscript, 1200. St. Catherine's Monastery, Sinai, Egypt",
        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest",
        "@type": "sc:Manifest",
        "json_mistake": [
            {
                "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/sequence/normal",
                "@type": "sc:Sequence",
                "canvases": [
                    {
                        "@type": "sc:Canvas",
                        "label": "Front Board Outside",
                        "@id": "http://test-iiif.library.ucla.edu/ark%3A%2F21198%2Fz14b44n8/manifest/canvas/hm957748",
                        "width": 5332,
                        "height": 7006,
                        "images": [],
                    },
                ],
            }
        ],
    },
)



from pathlib import Path
from typing import Hashable, Optional
from unittest.mock import Mock

import pytest
from pydantic import AnyUrl, ValidationError

import feed_sinai.sinai_types as st
from feed_sinai.sinai_json_importer import SinaiJsonImporter

IMPORTER = SinaiJsonImporter(base_path="tests/sinai/export_test")


class ExampleModel(st.BaseModel):
    children: "tuple[ExampleModel, ...]" = tuple()
    a: Optional[int] = None
    b: Optional[int] = None
    c: Optional[str] = None


class TestBaseModel:
    def test_convert(self) -> None:
        class A(st.BaseModel):
            a: int

        class B(A):
            b: int

        first = A(a=1)
        second = first.convert(B, b=2)
        assert second == B(a=1, b=2)

    class TestDeepGet:
        @pytest.fixture
        def obj(self) -> st.Date:
            return st.Date(
                value="sometime", iso=st.Iso(not_before="1980", not_after="2025")
            )

        def test_gets_by_type(self, obj: st.Date) -> None:
            assert "sometime" in obj.deep_get(cls=str)

        def test_gets_by_type_from_children(self, obj: st.Date) -> None:
            assert {"1980", "2025"} <= set(
                obj.deep_get("not_before", "not_after", cls=str)
            )

        def test_gets_by_name(self, obj: st.Date) -> None:
            assert set(obj.deep_get("not_after", cls=str)) == {"2025"}

        def test_ignores_by_name(self, obj: st.Date) -> None:
            assert "1980" not in obj.deep_get(cls=str, exclude=["not_before"])

        def test_gets_submodels(self, obj: st.Date) -> None:
            assert set(obj.deep_get(cls=st.Iso)) == {obj.iso}

        def test_gets_assoc_name_item(self) -> None:
            name = TestAssocNameItem.EPHREM.convert(st.AssocNameItemUnmerged)
            para = st.ParaItemUnmerged(
                type=st.ControlledTerm(id="t", label="T"),
                locus="locus",
                lang=[st.ControlledTerm(id="l", label="L")],
                assoc_name=[name],
            )
            assert set(para.deep_get(cls=st.AssocNameItemUnmerged)) == {name}

        def test_fib(self) -> None:
            test_obj = ExampleModel.model_validate(
                {
                    "a": 1,
                    "b": 1,
                    "children": [
                        {"a": 2, "b": 3, "children": [{"a": 5, "b": 8}]},
                        {"a": 13, "b": 21},
                        {"children": [{"b": 55, "c": "no"}], "b": 34, "c": "nope"},
                    ],
                }
            )

            assert set(test_obj.deep_get("a", "b", cls=int)) == {
                1,
                2,
                3,
                5,
                8,
                13,
                21,
                34,
                55,
            }


class TestControlledTerm:
    CONTROLLED_TERM = st.ControlledTerm(id="abc", label="123")
    MAN = st.ControlledTerm(id="man", label="Man")

    def test_happy_path(self) -> None:
        result = st.ControlledTerm.model_validate_json('{"id": "abc", "label": "123"}')
        assert result == self.CONTROLLED_TERM

    def test_hashable(self) -> None:
        result = st.ControlledTerm.model_validate_json('{"id": "abc", "label": "123"}')
        assert isinstance(result, Hashable)

    def test_extra_field(self) -> None:
        with pytest.raises(ValidationError):
            st.ControlledTerm.model_validate_json(
                '{"id": "abc", "label": "123"}, "other": "xyz"'
            )

    def test_missing_id(self) -> None:
        with pytest.raises(ValidationError):
            st.ControlledTerm.model_validate_json('{"label": "123"}')

    def test_missing_value(self) -> None:
        with pytest.raises(ValidationError):
            st.ControlledTerm.model_validate_json('{"id": "abc"}')

    def test_empty_id(self) -> None:
        with pytest.raises(ValidationError):
            st.ControlledTerm.model_validate_json('{"id": ", "label": "123"}')

    def test_empty_value(self) -> None:
        with pytest.raises(ValidationError):
            st.ControlledTerm.model_validate_json('{"id": "abc", "label": "}')


# class TestGender:
#     def test_man(self) -> None:
#         result = st.Gender({"id": "man", "label": "Man"})
#         assert result == st.Gender.man

#     def test_woman(self) -> None:
#         result = st.Gender({"id": "woman", "label": "Woman"})
#         assert result == st.Gender.woman

#     def test_other(self) -> None:
#         result = st.Gender({"id": "other", "label": "Other"})
#         assert result == st.Gender.other

#     def test_invalid_gender(self) -> None:
#         """With appologies, we are only tracking 'man', 'woman', and 'other"""
#         with pytest.raises(ValueError):
#             st.Gender({"id": "something", "label": "else"})

#     def test_as_value(self) -> None:
#         class TestModel(BaseModel):
#             gender: st.Gender
#         result = TestModel.model_validate_json('{"id": "other", "label": "Other"}')
#         assert result.gender == st.Gender.other


class TestIso:
    ISO = st.Iso(not_before="0010", not_after="0100")

    def test_good_iso(self) -> None:
        assert (
            st.Iso.model_validate_json('{"not_before": "0010", "not_after": "0100"}')
            == self.ISO
        )

    def test_hashable(self) -> None:
        assert hash(self.ISO)

    def test_no_notafter(self) -> None:
        result = st.Iso.model_validate_json('{"not_before": "0010"}')
        assert result.not_before == "0010"
        assert result.not_after is None

    def test_no_notbefore(self) -> None:
        with pytest.raises(ValidationError):
            st.Iso.model_validate_json('{"not_after": "0010"}')

    def test_optional_month_and_year(self) -> None:
        result = st.Iso.model_validate_json(
            '{"not_before": "2017-07", "not_after": "2025-07-23"}'
        )
        assert result.not_before == "2017-07"
        assert result.not_after == "2025-07-23"
        assert tuple(result.years()) == (
            2017,
            2018,
            2019,
            2020,
            2021,
            2022,
            2023,
            2024,
            2025,
        )

    def test_negative_year(self) -> None:
        assert (
            st.Iso.model_validate_json('{"not_before": "-0003"}').not_before == "-0003"
        )

    class TestYears:
        def test_returns_range(self) -> None:
            range_obj = TestIso.ISO.years()
            assert isinstance(range_obj, range)
            result = [*range_obj]
            assert result[:3] == [10, 11, 12] and result[-3:] == [98, 99, 100]

        def test_no_notafter(self) -> None:
            result = [*st.Iso.model_validate_json('{"not_before": "0010"}').years()]
            assert result == [10]

        def test_negative_and_zero_years(self) -> None:
            result = st.Iso.model_validate_json(
                '{"not_before": "-0003-04-05", "not_after": "0002"}'
            )
            assert [*result.years()] == [-3, -2, -1, 0, 1, 2]


class TestDate:
    DATE = st.Date(value="4th c. CE", iso=TestIso.ISO)

    def test_good_date(self) -> None:
        st.Date.model_validate_json(
            """
            {
                "value": "4th c. CE",
                "iso": {
                    "not_before": "0010",
                    "not_after": "0100"
                }
            }
        """
        )


class TestRelConItem:
    def test_good_json(self) -> None:
        result = st.RelConItem.model_validate_json(
            """
            {
                "label": "Onuphrius, Saint, -approximately 400",
                "uri": "https://w3id.org/haf/person/232371232899",
                "source": "HAF"
            }
        """
        )
        assert result.label == "Onuphrius, Saint, -approximately 400"
        assert result.source == st.RelatedConceptSource.HAF

    def test_invalid_source(self) -> None:
        with pytest.raises(ValidationError):
            st.RelConItem.model_validate_json(
                """
                {
                    "label": "Onuphrius, Saint, -approximately 400",
                    "uri": "https://w3id.org/haf/person/232371232899",
                    "source": "UCLA"
                }
            """
            )


class TestRefnoItem:
    def test_good_json(self) -> None:
        result = st.RefnoItem.model_validate_json(
            """
            {
                "label": "Homiliae super psalmos",
                "idno": "2836.001",
                "source": "CPG"
            }
        """
        )
        assert result.idno == "2836.001"


class TestBibItem:
    @pytest.fixture
    def json(self) -> str:
        return """
            {
                "id": "deb668b6-feec-4828-8749-a97441881226",
                "type": {
                    "id": "ref",
                    "label": "Reference"
                },
                "shortcode": "test shortcode",
                "citation": "test citation",
                "range": "[141], p. 156"
            }
        """

    def test_good_json(self, json: str) -> None:
        result = st.BibItem.model_validate_json(json)
        assert result.type.id == "ref"

    def test_hashable(self, json: str) -> None:
        result = st.BibItem.model_validate_json(json)
        assert hash(result)


# @pytest.mark.xfail
# class TestRelItem:
#     """Not currently used in the data"""
#     raise NotImplementedError


# @pytest.mark.xfail
# class TestRelAgentItem:
#     """Not currently used in the data"""
#     raise NotImplementedError


# @pytest.mark.xfail
# class TestRelPlaceItem:
#     """Not currently used in the data"""
#     raise NotImplementedError


# @pytest.mark.xfail
# class TestCataloguerItem:
#     """Not currently used in the data"""
#     raise NotImplementedError


class TestAgent:
    EPHREM = st.Agent(
        ark="ark:/21198/s1v887",
        type=st.ControlledTerm(id="person", label="Person"),
        pref_name="Ephrem",
        alt_name=["Ephrem the Syrian", "ܐܦܪܝܡ"],
        gender=TestControlledTerm.MAN,
        floruit=st.Date(
            value="303 CE-373 CE",
            iso=st.Iso(not_before="0303", not_after="0373"),
        ),
        rel_con=[
            st.RelConItem.model_validate(
                {
                    "label": "Ephraem, Syrus, Saint, 303-373",
                    "uri": "http://viaf.org/viaf/100177778",
                    "source": "VIAF",
                }
            ),
            st.RelConItem.model_validate(
                {
                    "label": "Ephraem, Syrus, Saint, 303-373",
                    "uri": "http://id.loc.gov/authorities/names/n50082928",
                    "source": "LoC",
                }
            ),
            st.RelConItem.model_validate(
                {
                    "label": "Ephrem, of Nisibis, 303-373",
                    "uri": "https://w3id.org/haf/person/818572788967",
                    "source": "HAF",
                }
            ),
            st.RelConItem.model_validate(
                {
                    "label": "Ephrem",
                    "uri": "http://syriaca.org/person/13",
                    "source": "Syriaca",
                }
            ),
            st.RelConItem.model_validate(
                {
                    "label": "Ephraem Graecus",
                    "uri": "https://pinakes.irht.cnrs.fr/notices/auteur/995/",
                    "source": "Pinakes",
                }
            ),
        ],
    )
    THEODORE = st.Agent(
        ark="ark:/21198/s1d01s",
        type=st.ControlledTerm(id="person", label="Person"),
        pref_name="Theodore the Studite",
        alt_name=["Theodore Studites"],
        gender=TestControlledTerm.MAN,
        floruit=st.Date.model_validate(
            {
                "value": "759 CE-826 CE",
                "iso": {"not_before": "0759", "not_after": "0826"},
            }
        ),
        rel_con=[
            st.RelConItem.model_validate(rc)
            for rc in [
                {
                    "label": "Theodore, Studites, Saint, 759-826",
                    "uri": "http://viaf.org/viaf/62875165",
                    "source": "VIAF",
                },
                {
                    "label": "Theodore, Studites, Saint, 759-826",
                    "uri": "https://id.loc.gov/authorities/names/n81118597",
                    "source": "LoC",
                },
                {
                    "label": "Theodore, Studites, 759-826",
                    "uri": "https://w3id.org/haf/person/636228715607",
                    "source": "HAF",
                },
                {
                    "label": "Theodorus Studita",
                    "uri": "https://pinakes.irht.cnrs.fr/notices/auteur/2685/",
                    "source": "Pinakes",
                },
            ]
        ],
    )

    def test_good_json(self) -> None:
        assert (
            st.Agent.model_validate_json(
                """
        {
            "ark": "ark:/21198/s1d01s",
            "type": {"id": "person", "label": "Person"},
            "pref_name": "Theodore the Studite",
            "alt_name": [
                "Theodore Studites"
            ],
            "gender": {"id": "man", "label": "Man"},
            "floruit": {
                "value": "759 CE-826 CE",
                "iso": {
                    "not_before": "0759",
                    "not_after": "0826"
                }
            },
            "rel_con": [
                {
                    "label": "Theodore, Studites, Saint, 759-826",
                    "uri": "http://viaf.org/viaf/62875165",
                    "source": "VIAF"
                },
                {
                    "label": "Theodore, Studites, Saint, 759-826",
                    "uri": "https://id.loc.gov/authorities/names/n81118597",
                    "source": "LoC"
                },
                {
                    "label": "Theodore, Studites, 759-826",
                    "uri": "https://w3id.org/haf/person/636228715607",
                    "source": "HAF"
                },
                {
                    "label": "Theodorus Studita",
                    "uri": "https://pinakes.irht.cnrs.fr/notices/auteur/2685/",
                    "source": "Pinakes"
                }
            ]
        }
        """
            )
            == self.THEODORE
        )

    def test_bad_ark(self) -> None:
        with pytest.raises(ValidationError, match="String should match pattern"):
            st.Agent.model_validate_json(
                """
            {
                "ark": "21198/s1d01s",
                "type": {"id": "person", "label": "Person"},
                "pref_name": "Theodore the Studite"
            }
            """
            )


class TestWritingItem:
    def test_good_writing_item(self) -> None:
        st.WritingItem.model_validate_json(
            """
            {
                "script": [
                    {
                        "id": "nuskhurimt",
                        "label": "Nuskhurimt",
                        "writing_system": "Georgian"
                    }
                ],
                "note": [
                    "Relatively thick and clumsy"
                ]
            }
        """
        )


class TestInkItem:
    def test_good_ink_item(self) -> None:
        st.InkItem.model_validate_json(
            """
            { 
                "locus": "ff. 55v-144v", 
                "color": [ 
                    "dark brown" 
                ], 
                "note": [ 
                    "Rubrication of titles in red ink" 
                ] 
            } 
        """
        )


class TestLayoutItem:
    def test_good_layout_item(self) -> None:
        st.LayoutItem.model_validate_json(
            """
            {
                "locus": "ff. 3r-54v",
                "columns": "1",
                "lines": "15",
                "dim": "Writing area: 215 x 140 mm",
                "note": [
                    "Possible pricking still visible in outer margins throughout",
                    "Text written inside bordered margins"
                ]
            }
        """
        )


class TestUnitItem:
    def test_good_text_unit_item(self) -> None:
        st.LayerTextUnit.model_validate_json(
            """
            {
                "id": "ark:/21198/s1w103",
                "label": "Item 1"
            }
        """
        )


class TestAssocNameItem:
    EPHREM = st.AssocNameItem(
        id="ark:/21198/s1v887",
        as_written="ܝܘܚܢܢ ܒܪܝ ܬܐܘܕܘܪܘܣ",
        role=st.ControlledTerm(id="scribe", label="Scribe"),
        note=["The ARK is for Ephrem, to demo functionality"],
    )

    def test_good_TestAssocNameItem(self) -> None:
        assert (
            st.AssocNameItem.model_validate_json(
                """
            { 
                "id": "ark:/21198/s1v887",
                "as_written": "ܝܘܚܢܢ ܒܪܝ ܬܐܘܕܘܪܘܣ", 
                "role": { 
                    "id": "scribe", 
                    "label": "Scribe" 
                }, 
                "note": [ 
                    "The ARK is for Ephrem, to demo functionality" 
                ] 
            }
        """
            )
            == self.EPHREM
        )

    def test_hashable(self) -> None:
        assert isinstance(self.EPHREM, Hashable)
        hash(self.EPHREM)

    def test_good_TestAssocNameItemMerged(self) -> None:
        result = self.EPHREM.convert(
            st.AssocNameItemMerged, agent_record=TestAgent.EPHREM
        )
        assert result.agent_record == TestAgent.EPHREM


class TestAssocPlaceItem:
    def test_good_assoc_place_item(self) -> None:
        st.AssocPlaceItem.model_validate_json(
            """
            {
                "value": "Possibly Jerusalem",
                "event": {
                    "id": "origin",
                    "label": "Place of Origin"
                }
            }
        """
        )


class TestAssocDateItem:
    def test_good_assoc_date_item(self) -> None:
        st.AssocDateItem.model_validate_json(
            """
            { 
                "type": { 
                    "id": "origin", 
                    "label": "Origin Date" 
                }, 
                "note": [ 
                    "Paleographic dating" 
                ], 
                "value": "Second half 9th c. CE", 
                "iso": { 
                    "not_before": "0851", 
                    "not_after": "0900" 
                } 
            }
        """
        )


class TestParaItem:
    def test_good_para_item(self) -> None:
        st.ParaItem.model_validate_json(
            """
            {
                "type": {
                    "id": "colophon",
                    "label": "Colophon"
                },
                "locus": "99r",
                "lang": [
                    {
                        "id": "class1252",
                        "label": "Syriac"
                    }
                ],
                "as_written": "ܐܫܠܡ ܒܥܘܕܪܢ ܐܠܗܐ ܡܢܘ ܕܬܝܒ̄ܘ ܥܠ ܐܦܝ ܬܫܥ ܫ̈ܥܝܢ ܝܘܡ ܕܬܪܝܐ ܒ̄ܛ ܒܐܕܐܪ ܫܢܬ ܐܠܦ ܘܚܡܫ̄ ܡ̄ܘ ܡ̣ܢ ܕܐܠܣܟܢ̄ܕܪ ܟܬܒܗ̣ ܕܝܢ ܐܢܫ ܡܚܲܝܠܐ ܘܚܛܝܐ ܘܒܨ̇ܝܪܐ ܕܟܠܗܘܢ ܒܢ̈ܝܗܫܐ ܦܝܡ ܣ ܠܡܬ ܥܡ ܨܐܕܝܬ̤ ܒܪ ܨ̇ܗܘܝ ܡ̣ܢ ܩܪܝܬ̤ ܡܒܪܟܬܐ ܘܪܚܡܬ̤ ܠܡܫܝ̣ܚܐ ܨܕܐܢܝ ܡܢܝܚ ܡܪܝܐ ܢܦܫܗ̣ ܥܡ ܟܠܗܘܢ ܩܕܝܫܘ̈ܗܝ ܐܡܝܢ"
            }
        """
        )


class TestRelatedMs:
    def test_good_related_ms(self) -> None:
        st.RelatedMs.model_validate_json(
            """
            { 
                "type": { 
                    "id": "filiation", 
                    "label": "Filiation" 
                }, 
                "label": "Copied from Syriac 10", 
                "mss": [
                    { 
                        "label": "Sinai Syriac 10", 
                        "id": "ark:/21198/z1p57n0b" 
                    } 
                ], 
                "note": [ 
                    "This is a dummy note for dev purposes" ,
                    "Otherwise largely lost"
                ] 
            } 
        """
        )

    def test_no_ms_id(self) -> None:
        st.RelatedMs.model_validate_json(
            """
            { 
                "type": { 
                    "id": "disjecta", 
                    "label": "Disjecta Membra" 
                }, 
                "label": "Disjecta Membra from the final quire", 
                "note": [ 
                    "The last 4 folios are today Biblioteca Ambrosiana, A 296 inf., ff. 70–73 = Chabot 20 (4 ff.) + Mingana Syr. 632 (1 f.)" ,
                    "Identification of disjecta membra by Rossetto"
                ], 
                "mss": [
                    { 
                        "label": "Biblioteca Ambrosiana, A 296 inf., ff. 70–73 = Chabot 20 (4 ff.)", 
                        "url": "https://archive.org/details/ChabotInventaireDesFragmentsDeMssSyriaquesConservesALaBibliothequeAmbrosienneAMilan/page/n3/mode/2up" 
                    }, 
                    { 
                        "label": "Mingana Syr. 632 (1 f.)", 
                        "url": "http://epapers.bham.ac.uk/160" 
                    } 
                ] 
            }
        """
        )


class TestNoteItem:
    def test_good_note_item(self) -> None:
        """ "Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/layers/s18d1p.json"""

        st.NoteItem.model_validate_json(
            """
            {
                "type": {
                    "id": "general",
                    "label": "Other Notes"
                },
                "value": "Many additions and erasures in overtext"
            }
        """
        )


class TestInscribedLayer:
    def test_good_inscribed_layer(self) -> None:
        """ "Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/layers/s18d1p.json"""

        st.InscribedLayer.model_validate_json(
            """
            {
                "ark": "ark:/21198/s18d1p",
                "reconstruction": false,
                "label": "Sinai Georgian 34, Overtext (Undetermined (Georgian))",
                "state": {
                    "id": "overtext",
                    "label": "Overtext"
                },
                "writing": [
                    {
                        "script": [
                            {
                                "id": "georgian-undetermined",
                                "label": "Undetermined (Georgian)",
                                "writing_system": "Georgian"
                            }
                        ]
                    }
                ],
                "ink": [
                    {
                        "color": [
                            "black",
                            "red"
                        ],
                        "note": [
                            "Red used for rubrics"
                        ]
                    }
                ],
                "text_unit": [
                    {
                        "id": "ark:/21198/s1308n",
                        "label": "Item 1"
                    }
                ],
                "features": [
                    {
                        "id": "dated",
                        "label": "Dated"
                    }
                ],
                "assoc_date": [
                    {
                        "type": {
                            "id": "origin",
                            "label": "Date of Origin"
                        },
                        "value": "932  CE",
                        "iso": {
                            "not_before": "0932"
                        }
                    }
                ],
                "note": [
                    {
                        "type": {
                            "id": "general",
                            "label": "Other Notes"
                        },
                        "value": "Many additions and erasures in overtext"
                    }
                ],
                "parent": [
                    "ark:/21198/z1h13zxq"
                ]
            }
        """
        )


class TestManuscriptLayer:
    def test_good_ManuscriptLayerUnmerged(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        st.ManuscriptLayerUnmerged.model_validate_json(
            """
            { 
                "id": "ark:/21198/te5fp1ol", 
                "label": "Overtext layer (13th c., Melkite)", 
                "type": { 
                    "id": "overtext", 
                    "label": "Overtext" 
                }, 
                "locus": "ff. 3r-54v" 
            } 
        """
        )

    def test_ManuscriptLayerMerged(self) -> None:
        st.ManuscriptLayerMerged(
            id="ark:/21198/123",
            label="Test Layer",
            type=st.ControlledTerm(id="guest", label="Guest Content"),
            layer_record=Mock(st.InscribedLayerMerged),
        )

    def test_good_UndertextManuscriptLayerMerged(self) -> None:
        result = st.UndertextManuscriptLayerMerged(
            id="ark:/21198/123",
            label="Test Layer",
            script=["Tengwar"],
            lang=["Sindarin"],
            type=st.ControlledTerm(id="undertext", label="Undertext"),
        ).model_dump()
        assert result == {
            "uto_layer_ark": "ark:/21198/123",
            "label": "Test Layer",
            "script": ("Tengwar",),
            "lang": ("Sindarin",),
            "type": {"id": "undertext", "label": "Undertext"},
        }

    def test_UndertextManuscriptLayerMerged_with_layer_record(self) -> None:
        with pytest.raises(
            ValidationError, match=r"layer_record\s+Input should be None"
        ):
            st.UndertextManuscriptLayerMerged(
                id="ark:/21198/123",
                label="Test Layer",
                layer_record=Mock(st.InscribedLayerMerged),
                type=st.ControlledTerm(id="undertext", label="Undertext"),
            )

    def test_UndertextManuscriptLayerMerged_wrong_type(self) -> None:
        with pytest.raises(
            ValidationError,
            match='UndertextManuscriptLayerMerged records must have type="undertext"',
        ):
            st.UndertextManuscriptLayerMerged(
                id="ark:/21198/123",
                label="Test Layer",
                type=st.ControlledTerm(id="guest", label="Guest Content"),
            )


class TestPart:
    def test_good_Part(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        st.Part.model_validate_json(
            """
            { 
                "label": "Part 1", 
                "summary": "Synaxarion (Gospel Lectionary for the movable feast days according to the Byzantine rite)", 
                "locus": "ff. 3-54", 
                "support": [
                    { 
                        "id": "paper", 
                        "label": "Paper" 
                    } 
                ], 
                "extent": "51 ff.", 
                "dim": "235 x 154 mm (average folio)", 
                "layer": [
                    { 
                        "id": "ark:/21198/te5fp1ol", 
                        "label": "Overtext layer (13th c., Melkite)", 
                        "type": { 
                            "id": "overtext", 
                            "label": "Overtext" 
                        }, 
                        "locus": "ff. 3r-54v" 
                    } 
                ], 
                "note": [
                    { 
                        "type": 
                        { 
                            "id": "support", 
                            "label": "Support" 
                        }, 
                        "value": "Oriental paper" 
                    }, 
                    { 
                        "type": 
                        { 
                            "id": "collation", 
                            "label": "Collation" 
                        }, 
                        "value": "Quire signatures in the first part are marked in the bottom margin on the recto side of the first folio of a quire; quires of 8 ff. with the exception of quire II (6 ff.)" 
                    }, 
                    { 
                        "type": 
                        { 
                            "id": "collation", 
                            "label": "Collation" 
                        }, 
                        "value": "F. 11 may be a replacement" 
                    }, 
                    { 
                        "type": 
                        { 
                            "id": "collation", 
                            "label": "Collation" 
                        }, 
                        "value": "One f. missing after f. 17" 
                    } 
                ] 
            }
        """
        )


class TestLocationItem:
    def test_good_LocationItem(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        st.LocationItem.model_validate_json(
            """
            { 
                "id": "sinai-oc", 
                "collection": "Old Collection" ,
                "repository": "St. Catherine's Monastery of the Sinai"
            }
        """
        )


class TestViscodexItem:
    def test_good_ViscodexItem(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        st.ViscodexItem.model_validate_json(
            """
            { 
                "type": { 
                    "id": "manuscript", 
                    "label": "Manuscript" 
                }, 
                "label": "Viscodex for Syriac 12", 
                "url": "https://vceditor.library.upenn.edu/project/668da6f75d69680001457684/viewOnly" 
            }
        """
        )


class TestIiifItem:
    def test_good_IiifItem(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        st.IiifItem.model_validate_json(
            """
            { 
                "type": { 
                    "id": "main", 
                    "label": "Main" 
                }, 
                "manifest": "https://iiif.library.ucla.edu/ark%3A%2F21198%2Fz15f0f9b/manifest", 
                "text_direction": "right-to-left", 
                "behavior": "paged", 
                "thumbnail": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz15f0f9b%2Fp161m45m/full/!200,200/0/default.jpg" 
            }
        """
        )

    def test_ingest_in_url(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        result = st.IiifItem.model_validate_json(
            """
            { 
                "type": { 
                    "id": "main", 
                    "label": "Main" 
                }, 
                "manifest": "https://ingest.iiif.library.ucla.edu/ark%3A%2F21198%2Fz15f0f9b/manifest", 
                "text_direction": "right-to-left", 
                "behavior": "paged", 
                "thumbnail": "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz15f0f9b%2Fp161m45m/full/!200,200/0/default.jpg" 
            }
        """
        )
        assert result.manifest == AnyUrl(
            "https://iiif.library.ucla.edu/ark%3A%2F21198%2Fz15f0f9b/manifest"
        )


class TestManuscriptObject:
    def test_good_ManuscriptObject(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/ms_objs/te5f0f9b.json"""

        result = st.ManuscriptObject.model_validate_json(
            Path("tests/sinai/export_test/ms_objs/te5f0f9b.json").read_text()
        )
        assert (
            result.image_provenance.program[0].camera_operator[0]  # type: ignore
            == "Damianos Kasotakis"
        )

    def test_unmerged_reconstructed_from_is_ark(self) -> None:
        result = st.ManuscriptObjectUnmerged(
            ark="ark:/21198/test",
            reconstruction=True,
            type=Mock(st.ControlledTerm),
            shelfmark="abc",
            state=Mock(st.ControlledTerm),
            part=(Mock(st.PartUnmerged),),
            location=(Mock(st.LocationItem),),
            reconstructed_from=("ark:/21198/456", "ark:/21198/789"),
        )
        assert result.reconstructed_from == ("ark:/21198/456", "ark:/21198/789")

    def test_unmerged_reconstructed_from_not_object(self) -> None:
        with pytest.raises(ValidationError):
            st.ManuscriptObjectUnmerged(
                ark="ark:/21198/test",
                reconstruction=True,
                type=Mock(st.ControlledTerm),
                shelfmark="abc",
                state=Mock(st.ControlledTerm),
                part=(Mock(st.PartUnmerged),),
                location=(Mock(st.LocationItem),),
                reconstructed_from=(
                    Mock(st.ReconstructedFrom),
                    Mock(st.ReconstructedFrom),
                ),
            )

    def test_merged_reconstructed_from_is_object(self) -> None:
        result = st.ManuscriptObjectMerged(
            ark="ark:/21198/test",
            reconstruction=True,
            type=Mock(st.ControlledTerm),
            shelfmark="abc",
            state=Mock(st.ControlledTerm),
            part=(Mock(st.PartMerged),),
            location=(Mock(st.LocationItem),),
            reconstructed_from=(Mock(st.ReconstructedFrom), Mock(st.ReconstructedFrom)),
        )
        assert isinstance(result.reconstructed_from[0], st.ReconstructedFrom)

    def test_merged_reconstructed_from_not_ark(self) -> None:
        with pytest.raises(ValidationError):
            st.ManuscriptObjectMerged(
                ark="ark:/21198/test",
                reconstruction=True,
                type=Mock(st.ControlledTerm),
                shelfmark="abc",
                state=Mock(st.ControlledTerm),
                part=(Mock(st.PartMerged),),
                location=(Mock(st.LocationItem),),
                reconstructed_from=("ark:/21198/456", "ark:/21198/789"),
            )


class TestWorkStub:
    def test_good_WorkStub(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/d58a6555cf08448fa0b0b95e0ff91717992d67f2/text_units/s1cd6h.json"""
        st.WorkStub.model_validate_json(
            """
            {
                "id": "ark:/21198/s12c7r"
            }
        """
        )


class TestWorkBrief:
    def test_good_WorkBrief(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/626274aac4d5f9004db44615827ebc167da00036/export_test/text_units/s1mh4z.json"""
        st.WorkBrief.model_validate_json(
            """
            {
                "desc_title": "Unidentified text"
            }
        """
        )


class TestExcerptItem:
    def test_good_ExcerptItem(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/46584bb30e5f351e73010a4914a71a52e3ea389d/text_units/s1b35q.json"""
        st.ExcerptItem.model_validate_json(
            """
            {
                "type": {
                    "id": "incipit",
                    "label": "Incipit"
                },
                "locus": "f. 8v",
                "as_written": "هذا مبتداء"
            }
        """
        )


class TestContents:
    def test_good_Contents(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/46584bb30e5f351e73010a4914a71a52e3ea389d/text_units/s1c37s.json"""
        st.Contents.model_validate_json(
            """
            {
                "label": "Gospel of Matthew (ff. 3v–39v)"
            }
        """
        )


class TestWorkWitItem:
    def test_good_WorkWitItem_with_WorkStub(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/d58a6555cf08448fa0b0b95e0ff91717992d67f2/text_units/s1cd6h.json"""
        result = st.WorkWitItem.model_validate_json(
            """
            {
                "work": {
                    "id": "ark:/21198/s12c7r"
                }
            }
        """
        )
        assert isinstance(result.work, st.WorkStub)

    def test_good_WorkWitItem_with_WorkBrief(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/d58a6555cf08448fa0b0b95e0ff91717992d67f2/text_units/s1cd6h.json"""
        result = st.WorkWitItem.model_validate_json(
            """
            {
                "work": {
                    "desc_title": "Gospel of Matthew (ff. 3v–39v)"
                }
            }
        """
        )
        assert isinstance(result.work, st.WorkBrief)

    def test_bad_mixed_work_type(self) -> None:
        with pytest.raises(ValidationError, match="Extra inputs are not permitted"):
            st.WorkWitItem.model_validate_json(
                """
                {
                    "work": {
                        "id": "ark:/21198/s12c7r",
                        "label": "Gospel of Matthew (ff. 3v–39v)"
                    }
                }
            """
            )


class TestTextUnit:
    def test_good_TextUnit(self) -> None:
        """example from https://github.com/UCLALibrary/sinaiportal_data/blob/d58a6555cf08448fa0b0b95e0ff91717992d67f2/text_units/s1cd6h.json"""
        st.TextUnit.model_validate_json(
            """
            {
                "ark": "ark:/21198/s1cd6h",
                "reconstruction": false,
                "label": "Gospel of Matthew, chs. 15-27",
                "lang": [
                    {
                        "id": "class1252",
                        "label": "Syriac"
                    }
                ],
                "work_wit": [
                    {
                        "work": {
                            "id": "ark:/21198/s12c7r"
                        }
                    }
                ],
                "note": [
                    {
                        "type": {
                            "id": "contents",
                            "label": "Contents Note"
                        },
                        "value": "Chs. 15-27"
                    },
                    {
                        "type": {
                            "id": "general",
                            "label": "Other Notes"
                        },
                        "value": "Description by Grigory Kessel"
                    }
                ],
                "parent": [
                    "ark:/21198/s1qd1q"
                ]
            }    
        """
        )


class TestCreation:
    def test_good_Creation(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/32278a93089ee067da48dac6adee5bb1ef888986/export_test/works/s1cc8x.json"""
        st.Creation.model_validate_json(
            """
            {
                "value": "600 CE",
                "iso": {
                    "not_before": "0600"
                }
            }
        """
        )


# # TODO not attested
# class TestIncipit:
#     def test_good_Incipit(self) -> None:
#         st.Incipit.model_validate_json("""

#         """)


# # TODO not attested
# class TestExplicit:
#     def test_good_Explicit(self) -> None:
#         st.Explicit.model_validate_json("""

#         """)


# # TODO not attested
# class TestRelWorkItem:
#     def test_good_RelWorkItem(self) -> None:
#         st.RelWorkItem.model_validate_json("""

#         """)


class TestConceptualWorkUnmerged:
    def test_good_ConceptualWorkUnmerged(self) -> None:
        """Example from https://github.com/UCLALibrary/sinaiportal_data/blob/32278a93089ee067da48dac6adee5bb1ef888986/export_test/works/s1cs35.json"""
        st.ConceptualWork.model_validate_json(
            """
            {
                "ark": "ark:/21198/s1cs35",
                "pref_title": "Ezekiel",
                "alt_title": [
                    "Bible. Ezekiel"
                ],
                "genre": [
                    {
                        "id": "biblical-texts",
                        "label": "Biblical texts"
                    }
                ],
                "rel_con": [
                    {
                        "label": "Bible. Ezekiel",
                        "uri": "https://viaf.org/viaf/176193843",
                        "source": "VIAF"
                    },
                    {
                        "label": "Bible. Ezekiel",
                        "uri": "http://id.loc.gov/authorities/names/n79117856",
                        "source": "LoC"
                    }
                ],
                "refno": [],
                "bib": []
            }
        """
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Convert UCLA Library CSV files for Ursus, our Blacklight installation."""

import asyncio
import importlib.metadata

import click

from feed_sinai.sinai_json_importer import SinaiJsonImporter


@click.group()
@click.version_option(version=importlib.metadata.version("feed_ursus"))
def sinai() -> None:
    """CLI for managing a Solr index for Ursus."""
    pass


@sinai.command("export")
@click.argument(
    "base_path",
    nargs=1,
    default=".",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
def export(base_path: str) -> None:
    importer = SinaiJsonImporter(base_path=base_path)
    importer.save_merged_records()


@sinai.command("load")
@click.argument(
    "base_path",
    nargs=1,
    default=".",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
@click.argument(
    "solr_url",
    nargs=1,
    default="http://localhost:8983/solr/californica",
    # help="URL of a solr instance, e.g. http://localhost:8983/solr/californica",
)
def load(base_path: str, solr_url: str) -> None:
    importer = SinaiJsonImporter(base_path=base_path, solr_url=solr_url)
    asyncio.run(importer.load_to_solr())


@sinai.command("wipe")
@click.argument(
    "solr_url",
    nargs=1,
    default="http://localhost:8983/solr/californica",
    # help="URL of a solr instance, e.g. http://localhost:8983/solr/californica",
)
def wipe(solr_url: str) -> None:
    importer = SinaiJsonImporter(solr_url=solr_url)
    importer.wipe_solr_records()


@sinai.command("save_solr_records")
@click.argument(
    "base_path",
    nargs=1,
    default=".",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
)
def save_solr_records(base_path: str) -> None:
    importer = SinaiJsonImporter(base_path=base_path)
    importer.save_solr_records()


if __name__ == "__main__":
    sinai()  # pylint: disable=no-value-for-parameter



# mypy: disallow_untyped_defs=False
"""Tests for feed_ursus.py"""

# pylint: disable=no-self-use

import os

from click.testing import CliRunner
from pysolr import Solr  # type: ignore

from feed_ursus import feed_ursus

SOLR_URL = os.getenv("SOLR_URL", "http://localhost:8983/solr/californica")


def test_feed_ursus():
    """Integration test for feed_ursus."""
    solr = Solr(SOLR_URL)

    # these should have been deleted already, but let's be sure
    solr.delete(id="xp6xn100zz-89112", commit=True)
    solr.delete(id="82765200zz-89112", commit=True)

    runner = CliRunner()
    result = runner.invoke(
        feed_ursus.feed_ursus,
        [
            "--solr_url",
            SOLR_URL,
            "load",
            "tests/csv/anais_collection.csv",
            "tests/csv/anais_work_simple.csv",
        ],
    )
    assert result.exit_code == 0

    collection_record = solr.search("id:xp6xn100zz-89112", defType="lucene")
    # Doesn't run against a fresh solr index, so there's no guarantee the result comes from this run of the feed_ursus command.
    # But at least we can see that pysolr works and talks to solr in this environment.
    assert collection_record.docs[0]["title_tesim"] == [
        "Nin (Anais) Papers, circa 1910-1977"
    ]

    work_record = solr.search("id:82765200zz-89112", defType="lucene").docs[0]
    assert work_record["title_tesim"] == ["Nin, Joaquin. 1914 [photograph]"]
    assert work_record["member_of_collections_ssim"] == [
        "Nin (Anais) Papers, circa 1910-1977"
    ]
    assert work_record["member_of_collection_ids_ssim"] == ["xp6xn100zz-89112"]

    # Delete and confirm
    result = runner.invoke(
        feed_ursus.feed_ursus,
        [
            "--solr_url",
            SOLR_URL,
            "delete",
            "--yes",
            "tests/csv/anais_collection.csv",
            "tests/csv/anais_work_simple.csv",
        ],
    )
    assert result.exit_code == 0
    assert (
        len(solr.search("id:xp6xn100zz-89112", defType="lucene").docs) == 0
    )  # collection record
    assert (
        len(solr.search("id:82765200zz-89112", defType="lucene").docs) == 0
    )  # work record


def test_feed_ursus_async():
    """Integration test for feed_ursus."""
    solr = Solr(SOLR_URL)

    # these should have been deleted already, but let's be sure
    solr.delete(id="xp6xn100zz-89112", commit=True)
    solr.delete(id="82765200zz-89112", commit=True)

    runner = CliRunner()
    result = runner.invoke(
        feed_ursus.feed_ursus,
        [
            "--solr_url",
            SOLR_URL,
            "load",
            "--async",
            "tests/csv/anais_collection.csv",
            "tests/csv/anais_work_simple.csv",
        ],
    )
    assert result.exit_code == 0

    collection_record = solr.search("id:xp6xn100zz-89112", defType="lucene")
    # Doesn't run against a fresh solr index, so there's no guarantee the result comes from this run of the feed_ursus command.
    # But at least we can see that pysolr works and talks to solr in this environment.
    assert collection_record.docs[0]["title_tesim"] == [
        "Nin (Anais) Papers, circa 1910-1977"
    ]

    work_record = solr.search("id:82765200zz-89112", defType="lucene").docs[0]
    assert work_record["title_tesim"] == ["Nin, Joaquin. 1914 [photograph]"]
    assert work_record["member_of_collections_ssim"] == [
        "Nin (Anais) Papers, circa 1910-1977"
    ]
    assert work_record["member_of_collection_ids_ssim"] == ["xp6xn100zz-89112"]

    # Delete and confirm
    result = runner.invoke(
        feed_ursus.feed_ursus,
        [
            "--solr_url",
            SOLR_URL,
            "delete",
            "--yes",
            "tests/csv/anais_collection.csv",
            "tests/csv/anais_work_simple.csv",
        ],
    )
    assert result.exit_code == 0
    assert (
        len(solr.search("id:xp6xn100zz-89112", defType="lucene").docs) == 0
    )  # collection record
    assert (
        len(solr.search("id:82765200zz-89112", defType="lucene").docs) == 0
    )  # work record



# """Tests for feed_sinai.py"""

# # pylint: disable=no-self-use

# import importlib
# import os
# import pytest

# from click.testing import CliRunner

# from feed_sinai import feed_sinai


# SOLR_URL = os.getenv("SOLR_URL", "http://localhost:8983/solr/californica")

# @pytest.mark.xfail
# def test_feed_sinai():
#     """Integration test for feed_sinai."""
#     solr = None
#     solr.delete(id="xp6xn100zz-89112", commit=True)
#     solr.delete(id="82765200zz-89112", commit=True)

#     runner = CliRunner()
#     result = runner.invoke(
#         feed_sinai.sinai,
#         [
#             "--solr_url",
#             SOLR_URL,
#             "load",
#             "tests/csv/anais_collection.csv",
#             "tests/csv/anais_work_simple.csv",
#         ],
#     )
#     assert result.exit_code == 0

#     collection_record = solr.search("id:xp6xn100zz-89112", defType="lucene")
#     # Doesn't run against a fresh solr index, so there's no guarantee the result comes from this run of the feed_sinai command.
#     # But at least we can see that pysolr works and talks to solr in this environment.
#     assert collection_record.docs[0]["title_tesim"] == [
#         "Nin (Anais) Papers, circa 1910-1977"
#     ]

#     work_record = solr.search("id:82765200zz-89112", defType="lucene").docs[0]
#     assert work_record["title_tesim"] == ["Nin, Joaquin. 1914 [photograph]"]
#     assert work_record["member_of_collections_ssim"] == [
#         "Nin (Anais) Papers, circa 1910-1977"
#     ]
#     assert work_record["member_of_collection_ids_ssim"] == ["xp6xn100zz-89112"]



# pylint: disable=no-self-use

import json

import pytest

import feed_sinai.sinai_types as st
from feed_sinai.sinai_json_importer import SinaiJsonImporter
from feed_sinai.solr_record import ManuscriptSolrRecord
from tests.sinai import test_sinai_types

# feed_sinai.mapper = importlib.import_module("feed_sinai.mapper.dlp")

BASE_PATH = "tests/sinai/export_test"


@pytest.fixture
def importer() -> SinaiJsonImporter:
    return SinaiJsonImporter(base_path=BASE_PATH)


def test_get_filename(importer: SinaiJsonImporter) -> None:
    assert importer.get_filename("ark:/21198/z1h13zxq") == "z1h13zxq.json"


def test_get_agent(importer: SinaiJsonImporter) -> None:
    result = importer.get_agent("ark:/21198/s1b59x")
    assert isinstance(result, st.Agent)
    assert result.model_dump() == {
        "ark": "ark:/21198/s1b59x",
        "type": {"id": "person", "label": "Person"},
        "pref_name": "Onuphrius",
        "gender": {"id": "man", "label": "Man"},
        "death": {
            "value": "ca. 400 CE",
            "iso": {"not_before": "0375", "not_after": "0425"},
        },
        "rel_con": (
            {
                "label": "Onuphrius, Saint, -approximately 400",
                "uri": st.AnyUrl("http://viaf.org/viaf/20485021"),
                "source": st.RelatedConceptSource.VIAF,
            },
            {
                "label": "Onuphrius, Saint, -approximately 400",
                "uri": st.AnyUrl("https://id.loc.gov/authorities/names/n92113349"),
                "source": st.RelatedConceptSource.LoC,
            },
            {
                "label": "Onuphrius, Saint, -approximately 400",
                "uri": st.AnyUrl("https://w3id.org/haf/person/232371232899"),
                "source": st.RelatedConceptSource.HAF,
            },
            {
                "label": "Onuphrius anachoreta in Aegypto",
                "uri": st.AnyUrl("https://pinakes.irht.cnrs.fr/notices/saint/691/"),
                "source": st.RelatedConceptSource.Pinakes,
            },
        ),
    }


class TestGetPlace:
    @pytest.mark.parametrize(
        ("ark", "expected"),
        (
            (
                "ark:/21198/pl1234",
                {
                    "ark": "ark:/21198/pl1234",
                    "pref_name": "Nisibis",
                    "alt_name": ("ܢܨܝܒܝܢ", "Nusaybin", "Ṣōbā"),
                },
            ),
            ("ark:/21198/pl5678", {"ark": "ark:/21198/pl5678", "pref_name": "Amid"}),
        ),
    )
    def test_get_good_place(
        self, ark: str, expected: st.Place, importer: SinaiJsonImporter
    ) -> None:
        assert importer.get_place(ark).model_dump() == expected

    def test_loads_all_places(self, importer: SinaiJsonImporter) -> None:
        n_files = 0
        for path in (importer.base_path / "places").glob("*.json"):
            importer.get_place("ark:/21198/" + path.stem)
            n_files += 1

        assert n_files == 2


def test_get_assoc_name_item(importer: SinaiJsonImporter) -> None:
    unmerged = test_sinai_types.TestAssocNameItem.EPHREM.convert(
        st.AssocNameItemUnmerged
    )
    result = importer.get_assoc_name_item(unmerged)
    assert isinstance(result, st.AssocNameItemMerged)
    assert result.agent_record and result.agent_record.alt_name == (
        "Ephrem the Syrian",
        "ܐܦܪܝܡ",
    )


class TestGetWork:
    def test_good_get_work(self, importer: SinaiJsonImporter) -> None:
        stub = st.WorkStub(id="ark:/21198/s1b015")
        result = importer.get_conceptual_work(stub)
        assert isinstance(result, st.ConceptualWorkMerged)
        assert result.pref_title == "2 John"

    def test_loads_all_works(self, importer: SinaiJsonImporter) -> None:
        n_files = 0
        for path in (importer.base_path / "works").glob("*.json"):
            stub = st.WorkStub(id="ark:/21198/" + path.stem)
            importer.get_conceptual_work(stub)
            n_files += 1
        assert n_files == 129


def test_get_work_brief(importer: SinaiJsonImporter) -> None:
    raw = st.WorkBriefUnmerged(desc_title="Abc123", creator=["ark:/21198/s1b59x"])
    result = importer.get_work_brief(raw)
    assert isinstance(result, st.WorkBriefMerged)
    assert result.creator and isinstance(result.creator[0].agent_record, st.Agent)
    assert result.creator[0].agent_record.pref_name == "Onuphrius"


@pytest.mark.parametrize(
    ("raw", "pref_title"),
    (
        (
            st.ContentsUnmerged(
                label="1st Week: Saturday of <the week of> Rest",
                work_id="ark:/21198/s1xs34",
                locus="f. 4r",
            ),
            "Acts",
        ),
        (st.ContentsUnmerged(work_id="ark:/21198/s12g6d", locus="f. 5r, 6r"), "3 John"),
        (
            st.ContentsUnmerged(
                label="3rd Week: Sunday, Saturday",
                locus="f. 7v, 8v",
                note=["Sub-heading: Joseph of Arimathea"],
            ),
            None,
        ),
    ),
)
def test_get_contents_item(
    raw: st.ContentsUnmerged, pref_title: str, importer: SinaiJsonImporter
) -> None:
    result = importer.get_contents_item(raw)
    assert result.pref_title == pref_title
    assert result.work_id == raw.work_id
    assert result.label == raw.label
    assert result.locus == raw.locus
    assert result.note == raw.note


class TestGetWorkWit:
    def test_get_work_wit_with_stub(self, importer: SinaiJsonImporter) -> None:
        raw = st.WorkWitItemUnmerged(
            work=st.WorkStub(id="ark:/21198/s1b015"),
        )
        result = importer.get_work_wit(raw)
        assert isinstance(result, st.WorkWitItemMerged)
        assert isinstance(result.work, st.ConceptualWorkMerged)
        assert result.work.pref_title == "2 John"

    def test_get_work_wit_with_workbrief(self, importer: SinaiJsonImporter) -> None:
        raw = st.WorkWitItemUnmerged(
            work=st.WorkBriefUnmerged(
                desc_title="Test Work", creator=["ark:/21198/s1b59x"]
            )
        )
        result = importer.get_work_wit(raw)
        assert isinstance(result, st.WorkWitItemMerged)
        assert isinstance(result.work, st.WorkBriefMerged)
        assert result.work.creator and isinstance(
            result.work.creator[0].agent_record, st.Agent
        )
        assert result.work.creator[0].agent_record.pref_name == "Onuphrius"

    def test_gets_contents_item(self, importer: SinaiJsonImporter) -> None:
        raw = st.WorkWitItemUnmerged(
            work=st.WorkBriefUnmerged(
                desc_title="Test Work", creator=["ark:/21198/s1b59x"]
            ),
            contents=[
                st.ContentsUnmerged(
                    label="1st Week: Saturday of <the week of> Rest",
                    work_id="ark:/21198/s1xs34",
                    locus="f. 4r",
                )
            ],
        )

        result = importer.get_work_wit(raw)

        assert result.contents[0].model_dump() == {
            "pref_title": "Acts",
            "label": "1st Week: Saturday of <the week of> Rest",
            "work_id": "ark:/21198/s1xs34",
            "locus": "f. 4r",
        }


def test_get_para(importer: SinaiJsonImporter) -> None:
    ct = st.ControlledTerm(id="x", label="X")

    raw = st.ParaItemUnmerged(
        type=ct,
        locus="abc",
        lang=[ct],
        assoc_name=[st.AssocNameItemUnmerged(id="ark:/21198/s1c304", role=ct)],
        assoc_place=[st.AssocPlaceItemUnmerged(id="ark:/21198/pl1234", event=ct)],
    )

    result = importer.get_para(raw)

    assert result.assoc_name[0].agent_record
    assert result.assoc_name[0].agent_record.pref_name == "Mar Saba"

    assert result.assoc_place[0].place_record
    assert result.assoc_place[0].place_record.pref_name == "Nisibis"


class TestGetTextUnit:
    def test_good_text_unit(self, importer: SinaiJsonImporter) -> None:
        stub = st.LayerTextUnitUnmerged.model_validate_json(
            """
            {
                "id": "ark:/21198/s1308n",
                "label": "Item 1"
            }
        """
        )
        result = importer.get_layer_text_unit(stub)
        assert result.model_dump() == {
            "id": "ark:/21198/s1308n",
            "label": "Item 1",
            "text_unit_record": {
                "ark": "ark:/21198/s1308n",
                "label": "Liturgical collection",
                "lang": ({"id": "nucl1302", "label": "Georgian"},),
                "parent": ("ark:/21198/s18d1p",),
                "reconstruction": False,
                "work_wit": (
                    {
                        "work": {
                            "desc_title": "Liturgical collection",
                            "genre": (
                                {
                                    "id": "liturgical-texts",
                                    "label": "Liturgical texts",
                                },
                            ),
                        },
                    },
                ),
            },
        }

    def test_loads_all_text_units(self, importer: SinaiJsonImporter) -> None:
        n_files = 0
        for path in (importer.base_path / "text_units").glob("*.json"):
            stub = st.LayerTextUnitUnmerged(
                id="ark:/21198/" + path.stem, label="whatevs"
            )
            importer.get_layer_text_unit(stub)
            n_files += 1
        assert n_files == 15


class TestGetLayer:
    def test_good_layer(self, importer: SinaiJsonImporter) -> None:
        raw = st.ManuscriptLayerUnmerged.model_validate_json(
            """
            {
                "id": "ark:/21198/ten0p1ol",
                "label": "Overtext layer (late 9th c., Kufic)",
                "type": {
                    "id": "overtext",
                    "label": "Overtext"
                    },
                "locus": "ff. 128-143"
            }
        """
        )
        result = importer.get_layer(raw).model_dump()

        assert result == {
            "id": "ark:/21198/ten0p1ol",
            "label": "Overtext layer (late 9th c., Kufic)",
            "type": {"id": "overtext", "label": "Overtext"},
            "locus": "ff. 128-143",
            "layer_record": {
                "ark": "ark:/21198/ten0p1ol",
                "reconstruction": False,
                "state": {"id": "overtext", "label": "Overtext"},
                "label": "Arabic NF M 28, Part 1, Overtext",
                "locus": "ff. 128-143",
                "summary": "Gospels, late 9th c., Arabic (Kufic)",
                "extent": "16 ff.",
                "writing": (
                    {
                        "script": (
                            {
                                "id": "kufic",
                                "label": "Kufic",
                                "writing_system": "Arabic",
                            },
                        ),
                        "locus": "ff. 128-143",
                    },
                ),
                "ink": (
                    {
                        "locus": "ff. 128-143",
                        "note": ("Titles in red ink",),
                    },
                ),
                "text_unit": (
                    {
                        "id": "ark:/21198/ten0p1olt1",
                        "label": "Primary Text Unit 1",
                        "locus": "ff. 128-143",
                        "text_unit_record": {
                            "ark": "ark:/21198/ten0p1olt1",
                            "reconstruction": False,
                            "label": "Arabic Gospels",
                            "locus": "ff. 128r-143v",
                            "lang": ({"id": "arab1395", "label": "Arabic"},),
                            "work_wit": (
                                {
                                    "work": {
                                        "ark": "ark:/21198/s12c7r",
                                        "pref_title": "Matthew",
                                        "alt_title": ("Bible. Matthew",),
                                        "genre": (
                                            {
                                                "id": "biblical-texts",
                                                "label": "Biblical texts",
                                            },
                                            {
                                                "id": "gospel-books",
                                                "label": "Gospel books",
                                            },
                                        ),
                                        "rel_con": (
                                            {
                                                "label": "Bible. Matthew",
                                                "uri": st.AnyUrl(
                                                    st.AnyUrl(
                                                        "https://viaf.org/viaf/188427863"
                                                    )
                                                ),
                                                "source": st.RelatedConceptSource.VIAF,
                                            },
                                            {
                                                "label": "Bible. Matthew",
                                                "uri": st.AnyUrl(
                                                    st.AnyUrl(
                                                        "http://id.loc.gov/authorities/names/n79056834"
                                                    )
                                                ),
                                                "source": st.RelatedConceptSource.LoC,
                                            },
                                        ),
                                    },
                                    "locus": "ff. 128r-130",
                                },
                                {
                                    "work": {
                                        "ark": "ark:/21198/s1630k",
                                        "pref_title": "Mark",
                                        "alt_title": ("Bible. Mark",),
                                        "genre": (
                                            {
                                                "id": "biblical-texts",
                                                "label": "Biblical texts",
                                            },
                                            {
                                                "id": "gospel-books",
                                                "label": "Gospel books",
                                            },
                                        ),
                                        "rel_con": (
                                            {
                                                "label": "Bible. Mark",
                                                "uri": st.AnyUrl(
                                                    "https://viaf.org/viaf/179823714"
                                                ),
                                                "source": st.RelatedConceptSource.VIAF,
                                            },
                                            {
                                                "label": "Bible. Mark",
                                                "uri": st.AnyUrl(
                                                    "http://id.loc.gov/authorities/names/n78095773"
                                                ),
                                                "source": st.RelatedConceptSource.LoC,
                                            },
                                        ),
                                    },
                                    "locus": "ff. 130v-135r",
                                },
                                {
                                    "work": {
                                        "ark": "ark:/21198/s1k88r",
                                        "pref_title": "Luke",
                                        "alt_title": ("Bible. Luke",),
                                        "genre": (
                                            {
                                                "id": "biblical-texts",
                                                "label": "Biblical texts",
                                            },
                                            {
                                                "id": "gospel-books",
                                                "label": "Gospel books",
                                            },
                                        ),
                                        "rel_con": (
                                            {
                                                "label": "Bible. Luke",
                                                "uri": st.AnyUrl(
                                                    "http://viaf.org/viaf/257061095"
                                                ),
                                                "source": st.RelatedConceptSource.VIAF,
                                            },
                                        ),
                                    },
                                    "locus": "ff. 135r-140r",
                                },
                                {
                                    "work": {
                                        "ark": "ark:/21198/s1388d",
                                        "pref_title": "John",
                                        "alt_title": ("Bible. John",),
                                        "genre": (
                                            {
                                                "id": "biblical-texts",
                                                "label": "Biblical texts",
                                            },
                                            {
                                                "id": "gospel-books",
                                                "label": "Gospel books",
                                            },
                                        ),
                                        "rel_con": (
                                            {
                                                "label": "Bible. John",
                                                "uri": st.AnyUrl(
                                                    "https://viaf.org/viaf/57145910123927021804"
                                                ),
                                                "source": st.RelatedConceptSource.VIAF,
                                            },
                                            {
                                                "label": "Bible. John",
                                                "uri": st.AnyUrl(
                                                    "http://id.loc.gov/authorities/names/n79060414"
                                                ),
                                                "source": st.RelatedConceptSource.LoC,
                                            },
                                        ),
                                    },
                                    "locus": "ff. 140v-143v",
                                },
                            ),
                            "note": (
                                {
                                    "type": {
                                        "id": "contents",
                                        "label": "Contents Note",
                                    },
                                    "value": "The Gospels continue in Arabic NF M 8 and NF M 27",
                                },
                            ),
                            "desc_provenance": {
                                "program": (
                                    {
                                        "label": "Sinai Palimpests Project",
                                        "description": "Described as part of the Sinai Palimpsests Project (2006-2017). The Sinai Palimpsests Project was sponsored by St. Catherine’s Monastery of the Sinai in partnership with the Early Manuscripts Electronic Library and the UCLA Library, and with funding from Arcadia. The Project provides scholarly identification and description of the undertext objects in a subset of palimpsested manuscripts in the Sinai collection, with minimal metadata for the overtexts of the host manuscripts.",
                                    },
                                )
                            },
                            "parent": ("ark:/21198/ten0p1ol",),
                            "internal": (
                                "Test record, delete after development is complete",
                            ),
                        },
                    },
                ),
                "assoc_date": (
                    {
                        "type": {"id": "origin", "label": "Origin Date"},
                        "note": ("Paleographic dating",),
                        "value": "Second half 9th c. CE",
                        "iso": {"not_before": "0851", "not_after": "0900"},
                    },
                ),
                "note": (
                    {
                        "type": {"id": "ornamentation", "label": "Ornamentation"},
                        "value": "Decorative headpieces throughout",
                    },
                    {
                        "type": {"id": "condition", "label": "Condition"},
                        "value": "Several damaged folios were repaired and reinforced more recently",
                    },
                ),
                "bib": (
                    {
                        "id": st.UUID("36ac2d29-349f-496d-b4ea-aff4e605c4ba"),
                        "type": {"id": "ref", "label": "Reference Work"},
                        "range": "p. 48-90",
                    },
                ),
                "parent": ("ark:/21198/ten02zkr",),
                "desc_provenance": {
                    "program": (
                        {
                            "label": "Sinai Palimpests Project",
                            "description": "Described as part of the Sinai Palimpsests Project (2006-2017). The Sinai Palimpsests Project was sponsored by St. Catherine’s Monastery of the Sinai in partnership with the Early Manuscripts Electronic Library and the UCLA Library, and with funding from Arcadia. The Project provides scholarly identification and description of the undertext objects in a subset of palimpsested manuscripts in the Sinai collection, with minimal metadata for the overtexts of the host manuscripts.",
                        },
                    )
                },
                "internal": ("Test record for development purposes; please delete.",),
            },
        }

    def test_loads_assoc_name_agents(self, importer: SinaiJsonImporter) -> None:
        result = importer.get_layer(
            st.ManuscriptLayerUnmerged(
                id="ark:/21198/te5fp1ol",
                label="abc",
                type=st.ControlledTerm(id="a", label="b"),
            )
        )
        assert result.layer_record.assoc_name[0].agent_record.pref_name == "Ephrem"  # type: ignore

    def test_loads_all_layers(self, importer: SinaiJsonImporter) -> None:
        n_files = 0
        for path in (importer.base_path / "layers").glob("*.json"):
            stub = st.ManuscriptLayerUnmerged(
                id="ark:/21198/" + path.stem,
                label="whatevs",
                type={"id": "what", "label": "ever"},
            )
            importer.get_layer(stub)
            n_files += 1
        assert n_files == 15


class TestGetMergedManuscript:
    def test_good_manuscript(self, importer: SinaiJsonImporter) -> None:
        with open(f"{BASE_PATH}/outputs/z1h13zxq.json", encoding="utf-8") as f:
            expected = json.load(f)

        result = importer.get_merged_manuscript(
            importer.base_path / "ms_objs/z1h13zxq.json"
        )

        test_ms_layer = result.part[0].uto[0]

        assert test_ms_layer.id == "ark:/21198/s1gh06"
        assert isinstance(test_ms_layer, st.UndertextManuscriptLayerMerged)
        assert test_ms_layer.model_dump().get("layer_record") is None

        assert json.loads(result.model_dump_json()) == expected

    @pytest.mark.xfail  # No test data yet
    def test_loads_assoc_name_agents(self) -> None:
        raise NotImplementedError

    @pytest.mark.xfail  # Don't seem to have good data here yet
    def test_loads_all_manuscripts(self, importer: SinaiJsonImporter) -> None:
        n_files = 0
        for path in (importer.base_path / "ms_objs").glob("*.json"):
            importer.get_merged_manuscript(path)
            n_files += 1
        assert n_files == 15


def test_iterate_merged_records(importer: SinaiJsonImporter) -> None:
    for ms_obj in importer.iterate_merged_records():
        assert isinstance(ms_obj, st.ManuscriptObjectMerged)
        assert isinstance(ms_obj.model_dump_json(), str)


def test_solr_record_works_on_all(importer: SinaiJsonImporter) -> None:
    for ms_obj in importer.iterate_merged_records():
        solr = ManuscriptSolrRecord(ms_obj=ms_obj)
        for field in ManuscriptSolrRecord.model_computed_fields:
            getattr(solr, field)

        assert isinstance(importer.solr_record(ms_obj=ms_obj), dict)



# mypy: disallow_untyped_defs=False
"""
tests for year_parser.py

Creates a multi-valued 'year_isim' field by parsing 'normalized_date'.
"""

import datetime

from feed_ursus import date_parser


def test_iso_8601():
    """Parses an iso 8601 standard string"""
    test_date = datetime.datetime(1941, 10, 1)
    assert date_parser.get_dates(["1941-10-01"]) == [test_date]


def test_just_year():
    """Parses a bare year."""
    test_date = datetime.datetime(1953, 1, 1)
    assert date_parser.get_dates(["1953"]) == [test_date]


def test_year_and_month():
    """Parses YYYY-MM"""
    test_date = datetime.datetime(1953, 10, 1)
    assert date_parser.get_dates(["1953-10"]) == [test_date]


def test_multiple_dates():
    """Parses multiple input values into a list of outputs."""
    test_dates = [
        datetime.datetime(1941, 10, 1),
        datetime.datetime(1935, 1, 1),
        datetime.datetime(1945, 1, 1),
    ]
    assert date_parser.get_dates(["1941-10-01", "1935", "1945"]) == sorted(test_dates)


def test_empty():
    """Returns an empty list if given an empty input."""
    assert date_parser.get_dates([]) == []


def test_unparseable():
    """Doesn't return anything for unparseable values, but still parses other elements in input."""
    test_date = datetime.datetime(1953, 1, 1)
    assert date_parser.get_dates(["1953", "[between 1928-1939]"]) == [test_date]


def test_range():
    """Parses YYYY/YYYY into dates."""

    assert date_parser.get_dates(["1937/1939", "1942/1943"]) == [
        datetime.datetime(1937, 1, 1),
        datetime.datetime(1939, 1, 1),
        datetime.datetime(1942, 1, 1),
        datetime.datetime(1943, 1, 1),
    ]

    # Parses YYY/YYYY into dates.

    assert date_parser.get_dates(["990/1000"]) == [
        datetime.datetime(990, 1, 1),
        datetime.datetime(1000, 1, 1),
    ]


def test_range_with_months():
    """Months can be included in range elements, but are ingored."""

    assert date_parser.get_dates(["1934-06/1934-07"]) == [
        datetime.datetime(1934, 6, 1),
        datetime.datetime(1934, 7, 1),
    ]


def test_duplicates():
    """Deduplicates elements in output."""

    assert date_parser.get_dates(["1934-06/1935-07", "1934-06-01", "1934"]) == [
        datetime.datetime(1934, 1, 1),
        datetime.datetime(1934, 6, 1),
        datetime.datetime(1935, 7, 1),
    ]



# -*- coding: utf-8 -*-
"""
Import JSON data into https://sinaimanuscripts.library.ucla.edu/

Input files are stored in https://github.com/UCLALibrary/sinaiportal_data
Output is pushed to a solr index suitable for use by https://github.com/UCLALibrary/sinaimanuscripts
"""

import asyncio
import json
import logging
from math import inf
from pathlib import Path
from typing import Any, Awaitable, Iterator, Optional

import httpx
import rich.progress
from pysolr import Solr, SolrError  # type: ignore

import feed_sinai.sinai_types as st
from feed_sinai.solr_record import ManuscriptSolrRecord


class SinaiJsonImporter:
    """Importer class to map data from"""

    base_path: Path
    solr: Solr
    solr_url: str | None
    async_client = httpx.AsyncClient()
    connection_pool = asyncio.Semaphore(3)

    _ms_objs_merged: dict[Path, st.ManuscriptObjectMerged]

    def __init__(self, base_path: str = ".", solr_url: Optional[str] = None):
        self.base_path = Path(base_path)
        self.solr = Solr(solr_url, always_commit=True)
        self.solr_url = solr_url

        self._ms_objs_merged = dict()

    @staticmethod
    def get_filename(ark: str) -> str:
        """Returns a filename based on an item's ark.

        Drops "ark:/21198/" (all records are assigned the UCLA NAAN) and adds the ".json" suffix"
        """

        return ark.replace("ark:/21198/", "").replace("/", "-") + ".json"

    def get_agent(self, ark: str) -> st.Agent:
        path = self.base_path / "agents" / self.get_filename(ark)
        return st.Agent.model_validate_json(path.read_text())

    def get_place(self, ark: str) -> st.Place:
        path = self.base_path / "places" / self.get_filename(ark)
        return st.Place.model_validate_json(path.read_text())

    def get_assoc_place_item(
        self, raw: st.AssocPlaceItemUnmerged
    ) -> st.AssocPlaceItemMerged:
        return raw.convert(
            st.AssocPlaceItemMerged,
            place_record=self.get_place(raw.id) if raw.id else None,
        )

    def get_assoc_name_item(
        self, raw: st.AssocNameItemUnmerged
    ) -> st.AssocNameItemMerged:
        return raw.convert(
            st.AssocNameItemMerged,
            agent_record=self.get_agent(raw.id) if raw.id else None,
        )

    def get_conceptual_work(self, stub: st.WorkStub) -> st.ConceptualWorkMerged:
        path = self.base_path / "works" / self.get_filename(stub.id)
        raw = st.ConceptualWorkUnmerged.model_validate_json(path.read_text())

        return raw.convert(
            st.ConceptualWorkMerged,
            creator=[
                self.get_assoc_name_item(assoc_name) for assoc_name in raw.creator
            ],
        )

    def get_work_brief(self, raw: st.WorkBriefUnmerged) -> st.WorkBriefMerged:
        return raw.convert(
            st.WorkBriefMerged,
            creator=[
                st.WorkBriefCreator(id=id, agent_record=self.get_agent(id))
                for id in raw.creator
            ],
        )

    def get_contents_item(self, raw: st.ContentsUnmerged) -> st.ContentsMerged:
        if not raw.work_id:
            return raw.convert(st.ContentsMerged)

        work_record = self.get_conceptual_work(stub=st.WorkStub(id=raw.work_id))

        return raw.convert(st.ContentsMerged, pref_title=work_record.pref_title)

    def get_work_wit(self, raw: st.WorkWitItemUnmerged) -> st.WorkWitItemMerged:
        return raw.convert(
            st.WorkWitItemMerged,
            work=(
                self.get_work_brief(raw.work)
                if isinstance(raw.work, st.WorkBrief)
                else self.get_conceptual_work(raw.work)
            ),
            contents=[self.get_contents_item(raw_item) for raw_item in raw.contents],
        )

    def get_para(self, raw: st.ParaItemUnmerged) -> st.ParaItemMerged:
        return raw.convert(
            st.ParaItemMerged,
            assoc_name=[self.get_assoc_name_item(name) for name in raw.assoc_name],
            assoc_place=[self.get_assoc_place_item(place) for place in raw.assoc_place],
        )

    def get_text_unit(self, ark: st.Ark) -> st.TextUnitMerged:
        path = self.base_path / "text_units" / self.get_filename(ark)
        raw = st.TextUnitUnmerged.model_validate_json(path.read_text())

        return raw.convert(
            st.TextUnitMerged,
            work_wit=[self.get_work_wit(work_wit) for work_wit in raw.work_wit],
            para=[self.get_para(para) for para in raw.para],
            reconstructed_from=[
                st.ReconstructedFrom(
                    id=ark, shelfmark=self.get_merged_manuscript(ark).shelfmark
                )
                for ark in raw.reconstructed_from
            ],
        )

    def get_layer_text_unit(
        self, raw: st.LayerTextUnitUnmerged
    ) -> st.LayerTextUnitMerged:
        return raw.convert(
            st.LayerTextUnitMerged, text_unit_record=self.get_text_unit(raw.id)
        )

    def get_uto_ms_ark(self, layer_record: st.InscribedLayerMerged) -> st.Ark | None:
        arks: list[st.Ark] = []
        for parent_ark in layer_record.parent:
            parent_ms = st.ManuscriptObjectUnmerged.model_validate_json(
                (self.base_path / "ms_objs" / self.get_filename(parent_ark)).read_text()
            )
            if parent_ms.type.id == "uto":
                arks.append(parent_ark)

        if len(arks) == 0:
            return None

        if len(arks) > 1:
            logging.warning(
                f"Multiple values found for `uto_ms_ark` in {layer_record}, using the first and discarding the rest"
            )

        return arks[0]

    def get_layer(
        self, ms_layer: st.ManuscriptLayerUnmerged
    ) -> st.ManuscriptLayerMerged:
        assert ms_layer.type.id != "undertext"

        layer_record_path = self.base_path / "layers" / self.get_filename(ms_layer.id)
        raw = st.InscribedLayerUnmerged.model_validate_json(
            layer_record_path.read_text()
        )

        layer_record = raw.convert(
            st.InscribedLayerMerged,
            text_unit=[
                self.get_layer_text_unit(text_unit) for text_unit in raw.text_unit
            ],
            para=[self.get_para(para) for para in raw.para],
            assoc_name=[
                self.get_assoc_name_item(name_item) for name_item in raw.assoc_name
            ],
            assoc_place=[self.get_assoc_place_item(place) for place in raw.assoc_place],
            reconstructed_from=[
                st.ReconstructedFrom(
                    id=ark, shelfmark=self.get_merged_manuscript(ark).shelfmark
                )
                for ark in raw.reconstructed_from
            ],
        )

        return ms_layer.convert(st.ManuscriptLayerMerged, layer_record=layer_record)

    def get_uto(
        self, ms_layer: st.ManuscriptLayerUnmerged
    ) -> st.ManuscriptLayerMerged | st.UndertextManuscriptLayerMerged:
        assert ms_layer.type.id == "undertext"

        layer_record_path = self.base_path / "layers" / self.get_filename(ms_layer.id)
        raw = st.InscribedLayerUnmerged.model_validate_json(
            layer_record_path.read_text()
        )

        layer_record = raw.convert(
            st.InscribedLayerMerged,
            text_unit=[
                self.get_layer_text_unit(text_unit) for text_unit in raw.text_unit
            ],
            para=[self.get_para(para) for para in raw.para],
            assoc_name=[
                self.get_assoc_name_item(name_item) for name_item in raw.assoc_name
            ],
        )

        return ms_layer.convert(
            st.UndertextManuscriptLayerMerged,
            uto_ms_ark=self.get_uto_ms_ark(layer_record),
            script=[
                script.label
                for writing_item in layer_record.writing
                for script in writing_item.script
            ],
            lang=[
                lang.label
                for text_unit in layer_record.text_unit
                for lang in text_unit.text_unit_record.lang
            ],
            orig_date=[
                date
                for date in (layer_record.assoc_date or [])
                if date.type.id == "origin"
            ],
        )

    def get_part(self, raw: st.PartUnmerged) -> st.PartMerged:
        return raw.convert(
            st.PartMerged,
            layer=[],
            ot_layer=[
                self.get_layer(layer)
                for layer in raw.layer
                if layer.type.id == "overtext"
            ],
            guest_layer=[
                self.get_layer(layer) for layer in raw.layer if layer.type.id == "guest"
            ],
            uto=[
                self.get_uto(layer)
                for layer in raw.layer
                if layer.type.id == "undertext"
            ],
            para=[self.get_para(para) for para in raw.para],
        )

    def get_merged_manuscript(
        self, path_or_ark: Path | str
    ) -> st.ManuscriptObjectMerged:
        if isinstance(path_or_ark, str):
            path = (
                self.base_path
                / "ms_objs"
                / (path_or_ark.removeprefix("ark:/21198/") + ".json")
            )
        else:
            path = path_or_ark

        if path in self._ms_objs_merged:
            return self._ms_objs_merged[path]

        raw = st.ManuscriptObjectUnmerged.model_validate_json(path.read_text())

        self._ms_objs_merged[path] = raw.convert(
            st.ManuscriptObjectMerged,
            part=[self.get_part(stub) for stub in raw.part],
            layer=[],
            ot_layer=[
                self.get_layer(layer)
                for layer in raw.layer
                if layer.type.id == "overtext"
            ],
            guest_layer=[
                self.get_layer(layer) for layer in raw.layer if layer.type.id == "guest"
            ],
            uto=[
                self.get_uto(layer)
                for layer in raw.layer
                if layer.type.id == "undertext"
            ],
            assoc_name=[self.get_assoc_name_item(name) for name in raw.assoc_name],
            assoc_place=[self.get_assoc_place_item(place) for place in raw.assoc_place],
            para=[self.get_para(para) for para in raw.para],
            reconstructed_from=[
                st.ReconstructedFrom(
                    id=ark, shelfmark=self.get_merged_manuscript(ark).shelfmark
                )
                for ark in raw.reconstructed_from
            ],
        )

        return self._ms_objs_merged[path]

    def iterate_merged_records(self) -> Iterator[st.ManuscriptObjectMerged]:
        """Yield json records for manuscripts with other data embedded."""

        paths = tuple((self.base_path / "ms_objs").glob("*.json"))
        for path in rich.progress.track(paths):
            try:
                yield self.get_merged_manuscript(path)
            except Exception as e:
                logging.warning(f"Could not merge {path}: {e}")

    def save_merged_records(self) -> None:
        (self.base_path / "merged").mkdir(exist_ok=True)
        for record in self.iterate_merged_records():
            path = self.base_path / "merged" / self.get_filename(record.ark)
            path.write_text(record.model_dump_json(indent=2))

    def solr_record(self, ms_obj: st.ManuscriptObjectMerged) -> dict[str, Any]:
        return json.loads(ManuscriptSolrRecord(ms_obj=ms_obj).model_dump_json())

    async def load_to_solr(self, batch_size: float = inf) -> None:
        """
        Loads records to Solr in batches. `batch_size` should be a positive integer or `math.inf`.
        """
        batch: list[dict] = list()
        results: list[Awaitable[None]] = list()

        for record in self.iterate_merged_records():
            try:
                batch.append(self.solr_record(record))
            except Exception as e:
                logging.warning(f"could not generate solr document {record.ark}: {e}")

            if len(batch) > 100:
                results.append(self.add_batch(batch))
                batch = list()

        if len(batch) > 0:
            results.append(self.add_batch(batch))

        await asyncio.gather(*results)

    async def add_batch(self, batch: list[dict]) -> None:
        try:
            async with self.connection_pool:
                response = await self.async_client.post(
                    f"{self.solr_url}/update?commit=true", json=batch
                )

            if response.is_error:
                raise SolrError(response.json().get("error").get("msg"))

        except Exception as e:
            if len(batch) == 1:
                print(f"Error adding record {batch[0]['id']}: {e}")
            else:
                mid = int(len(batch) / 2)
                await asyncio.gather(
                    self.add_batch(batch[:mid]), self.add_batch(batch[mid:])
                )

    def save_solr_records(self) -> None:
        (self.base_path / "solr").mkdir(exist_ok=True)
        for record in self.iterate_merged_records():
            path = self.base_path / "solr" / self.get_filename(record.ark)
            path.write_text(
                ManuscriptSolrRecord(ms_obj=record).model_dump_json(indent=2)
            )

    def wipe_solr_records(self) -> None:
        self.solr.delete(q="*:*")



# mypy: disallow_untyped_defs=False
"""Tests for mapper/sinai.py"""

# pylint: disable=no-self-use

import random

import pytest  # type: ignore

import feed_ursus.mapper.sinai as mapper


class TestVisibility:
    """Tests for mapper.visibility"""

    @pytest.mark.parametrize(
        ["visibility", "expected_result"],
        [
            # Mapping values
            ("authenticated", "authenticated"),
            ("discovery", "sinai"),
            ("open", "open"),
            ("private", "restricted"),
            ("public", "open"),
            ("registered", "authenticated"),
            ("restricted", "restricted"),
            ("sinai", "sinai"),
            ("ucla", "authenticated"),
            # String cleanup
            (" open", "open"),
            ("open       ", "open"),
            ("OPEN", "open"),
        ],
    )
    def test_visibility_included(self, visibility, expected_result):
        """Mappings of the 'visibility' field if included in csv"""
        row = {
            "Visibility": visibility,
            "Item Status": random.choice(("Completed", "Incomplete")),  # ignored
        }

        assert mapper.visibility(row) == expected_result

    @pytest.mark.parametrize(
        ["item_status", "expected_result"],
        [
            ("Completed", "open"),
            ("Completed with minimal metadata", "open"),
            ("Incomplete", "restricted"),
            ("Anything", "restricted"),
            ("Else", "restricted"),
        ],
    )
    def test_visibility_inferred(self, item_status, expected_result):
        """Infer from from 'Item Status' if 'Visibility' not in csv."""

        row = {"Item Status": item_status}
        assert mapper.visibility(row) == expected_result



# mypy: disallow_untyped_defs=False
"""Tests for feed_ursus.py"""

# pylint: disable=no-self-use

from unittest.mock import Mock, call

import pytest  # type: ignore
from httpx import AsyncClient, Response
from pysolr import Solr  # type: ignore

import feed_ursus.importer
from feed_ursus.importer import Importer, collate_child_works, get_bare_field_name

from . import fixtures  # pylint: disable=wrong-import-order


@pytest.fixture
def importer() -> Importer:
    importer = Importer(solr_url="", mapper_name="dlp")
    importer.solr_client = Mock(Solr)
    importer.async_client = Mock(AsyncClient)

    def mock_post(url: str, json: list[dict]):
        response = Mock(Response)
        response.is_error = False
        return response

    importer.async_client.post.side_effect = mock_post  # type: ignore

    return importer


class TestLoadCsv:
    """Tests for function load_csv"""

    def test_file_exists(self, importer):
        """gets the contents of a CSV file"""

        importer.load_csv(filenames=["tests/csv/anais_collection.csv"], batch=True)
        importer.solr_client.add.assert_called_once()

    def test_file_does_not_exist(self, importer):
        """raises an error if file does not exist"""

        with pytest.raises(FileNotFoundError):
            importer.load_csv(filenames=["tests/fixtures/nonexistent.csv"], batch=True)


class TestLoadCsvAsync:
    """Tests for function load_csv"""

    @pytest.mark.asyncio
    async def test_file_exists(self, importer):
        """gets the contents of a CSV file"""

        await importer.load_csv_async(filenames=["tests/csv/anais_collection.csv"])
        importer.async_client.post.assert_called_once()

    @pytest.mark.asyncio
    async def test_file_does_not_exist(self, importer):
        """raises an error if file does not exist"""

        with pytest.raises(FileNotFoundError):
            await importer.load_csv_async(filenames=["tests/fixtures/nonexistent.csv"])


class TestAddToSolr:
    @pytest.mark.asyncio
    async def test_divides_batch_and_retries(self, importer: Importer, capfd):
        batch = [{"id": n} for n in range(4)]

        def mock_post(url: str, json: list[dict]):
            response = Mock(Response)
            if batch[2] in json:
                response.is_error = True
                response.json = lambda: {"error": {"msg": "abcxyz"}}
                return response
            else:
                response.is_error = False
                return response

        importer.async_client.post.side_effect = mock_post  # type: ignore

        await importer.add_to_solr(batch)

        importer.async_client.post.assert_has_calls(  # type: ignore
            [
                call("/update?commit=true", json=batch),
                call("/update?commit=true", json=batch[0:2]),
                call("/update?commit=true", json=batch[2:4]),
                call("/update?commit=true", json=[batch[2]]),
                call("/update?commit=true", json=[batch[3]]),
            ]
        )
        assert "Error adding record 2:" in capfd.readouterr().out


class TestMapFieldValue:
    """tests for function map_field_value"""

    def test_parses_array(self, monkeypatch, importer):
        """parses value to an array of strings separated by '|~|'"""

        monkeypatch.setitem(
            importer.mapper.FIELD_MAPPING, "test_ursus_field_tesim", "Test DLCS Field"
        )
        input_record = {"Test DLCS Field": "one|~|two|~|three"}
        result = importer.map_field_value(
            input_record,
            "test_ursus_field_tesim",
        )
        assert result == [
            "one",
            "two",
            "three",
        ]

    def test_calls_function(self, importer, monkeypatch):
        """If mapper defines a function map_[SOLR_NAME], calls that function."""
        # pylint: disable=no-member
        monkeypatch.setitem(
            importer.mapper.FIELD_MAPPING,
            "test_ursus_field_tesim",
            lambda x: "lkghsdh",
        )
        result = importer.map_field_value({}, "test_ursus_field_tesim")
        assert result == "lkghsdh"


def test_get_bare_field_name():
    """function get_bare_field_name"""

    assert (
        get_bare_field_name("human_readable_test_field_name_tesim") == "test_field_name"
    )


class TestMapRecord:
    """function map_record"""

    COLLECTION_NAMES = {"noitcelloc-321": "Test Collection KGSL"}

    def test_maps_record(self, importer, monkeypatch):
        """maps the record for Ursus"""
        monkeypatch.setattr(
            importer.mapper,
            "FIELD_MAPPING",
            {
                "id": lambda r: r["Item ARK"],
                "test_ursus_field_tesim": "Test DLCS Field",
            },
        )
        result = importer.map_record(
            {"Item ARK": "ark:/123/abc", "Test DLCS Field": "lasigd|~|asdfg"}
        )

        assert result.pop("ingest_id_ssi")
        assert result == {
            "features_sim": None,
            "genre_sim": None,
            "human_readable_language_sim": None,
            "human_readable_resource_type_sim": None,
            "id": "ark:/123/abc",
            "location_sim": None,
            "member_of_collections_ssim": [],
            "named_subject_sim": None,
            "place_of_origin_sim": None,
            "script_sim": None,
            "subject_sim": None,
            "support_sim": None,
            "test_ursus_field_tesim": ["lasigd", "asdfg"],
            "thumbnail_url_ss": None,
            "writing_system_sim": None,
            "year_isim": [],
            "names_sim": None,
            "architect_sim": None,
            "author_sim": None,
            "calligrapher_sim": None,
            "combined_subject_ssim": [],
            "commentator_sim": None,
            "composer_sim": None,
            "editor_sim": None,
            "engraver_sim": None,
            "illuminator_sim": None,
            "illustrator_sim": None,
            "lyricist_sim": None,
            "printmaker_sim": None,
            "rubricator_sim": None,
            "scribe_sim": None,
            "uniform_title_sim": None,
            "translator_sim": None,
            "associated_name_sim": None,
            "form_sim": None,
            "date_dtsim": [],
            "header_index_tesim": [],
            "name_fields_index_tesim": [],
            "keywords_tesim": [],
            "keywords_sim": [],
            "collection_sim": None,
            "record_origin_ssi": "feed_ursus",
        }

    def test_sets_id(self, importer):
        """sets 'id' to reversed ark"""
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
            }
        )
        assert result["id"] == "cba-321"

    def test_sets_thumbnail(self, importer):
        """sets a thumbnail URL"""
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "IIIF Access URL": "https://test.iiif.server/url",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
            },
        )
        assert (
            result["thumbnail_url_ss"]
            == "https://test.iiif.server/url/full/!200,200/0/default.jpg"
        )

    def test_sets_access(self, importer):
        """sets permissive values for blacklight-access-control"""
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "Visibility": "open",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
            },
        )
        assert result["discover_access_group_ssim"] == ["public"]
        assert result["read_access_group_ssim"] == ["public"]
        assert result["download_access_person_ssim"] == ["public"]

    def test_sets_iiif_manifest_url(self, importer):
        """sets a IIIF manifest URL based on the ARK"""
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
            }
        )
        assert (
            result["iiif_manifest_url_ssi"]
            == "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest"
        )

    def test_sets_collection(self, importer):
        """sets the collection name by using the collection row"""

        importer.collection_names = self.COLLECTION_NAMES
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "Parent ARK": "ark:/123/collection",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
            }
        )
        assert result["member_of_collections_ssim"] == ["Test Collection KGSL"]

    @pytest.mark.parametrize(
        ["column_name", "facet_field_name"],
        [
            ("Coverage.geographic", "location_sim"),
            ("Language", "human_readable_language_sim"),
            ("Name.subject", "named_subject_sim"),
            ("Subject", "subject_sim"),
            ("Type.genre", "genre_sim"),
            ("Type.typeOfResource", "human_readable_resource_type_sim"),
        ],
    )
    def test_sets_facet_fields(self, column_name, facet_field_name, importer):
        """Copies *_tesim to *_sim fields for facets"""
        value = "value aksjg"
        result = importer.map_record(
            {
                "Item ARK": "ark:/123/abc",
                "IIIF Manifest URL": "https://iiif.library.ucla.edu/ark%3A%2F123%2Fabc/manifest",
                column_name: value,
            }
        )
        assert result[facet_field_name] == [value]


class TestThumbnailFromChild:
    """Tests for feed_ursus.thumbnail_from_child."""

    def test_uses_title(self, importer):
        """Returns the thumbnail from child row 'f. 001r'"""

        importer.child_works = collate_child_works(
            {
                "ark:/work/1": {
                    "Object Type": "Work",
                    "Item ARK": "ark:/work/1",
                    "Parent ARK": "ark:/collection/1",
                    "Thumbnail URL": None,
                    "Title": None,
                },
                "ark:/child/2": {
                    "Object Type": "ChildWork",
                    "Item ARK": "ark:/child/2",
                    "Parent ARK": "ark:/work/1",
                    "Thumbnail URL": "/thumb2.jpg",
                    "Title": "f. 001v",
                },
                "ark:/child/1": {
                    "Object Type": "ChildWork",
                    "Item ARK": "ark:/child/1",
                    "Parent ARK": "ark:/work/1",
                    "Thumbnail URL": "/thumb1.jpg",
                    "Title": "f. 001r",
                },
            }
        )
        record = {"ark_ssi": "ark:/work/1"}

        result = importer.thumbnail_from_child(record)
        assert result == "/thumb1.jpg"

    def test_uses_mapper(self, importer):
        """Uses the mapper to generate a thumbnail from access_copy, if necessary"""

        importer.child_works = collate_child_works(
            {
                "ark:/work/1": {
                    "Item ARK": "ark:/work/1",
                    "Parent ARK": "ark:/collection/1",
                    "IIIF Access URL": None,
                    "Title": None,
                    "Object Type": "Work",
                },
                "ark:/child/1": {
                    "Item ARK": "ark:/child/1",
                    "Parent ARK": "ark:/work/1",
                    "IIIF Access URL": "http://iiif.url/123",
                    "Title": "f. 001r",
                    "Object Type": "ChildWork",
                },
            }
        )
        record = {"ark_ssi": "ark:/work/1"}

        result = importer.thumbnail_from_child(record)
        assert result == "http://iiif.url/123/full/!200,200/0/default.jpg"

    def test_defaults_to_first(self, importer):
        """Returns the thumbnail from first child row if it can't find 'f. 001r'"""
        importer.child_works = collate_child_works(
            {
                "ark:/work/1": {
                    "Item ARK": "ark:/work/1",
                    "Parent ARK": "ark:/collection/1",
                    "Thumbnail URL": None,
                    "Title": None,
                    "Object Type": "Work",
                },
                "ark:/child/2": {
                    "Item ARK": "ark:/child/2",
                    "Parent ARK": "ark:/work/1",
                    "Thumbnail URL": "/thumb2.jpg",
                    "Title": "f. 001v",
                    "Object Type": "ChildWork",
                },
                "ark:/child/1": {
                    "Item ARK": "ark:/child/1",
                    "Parent ARK": "ark:/work/1",
                    "Thumbnail URL": "/thumb1.jpg",
                    "Title": "f. 002r",
                    "Object Type": "ChildWork",
                },
            }
        )
        record = {"ark_ssi": "ark:/work/1"}

        result = importer.thumbnail_from_child(record)
        assert result == "/thumb2.jpg"

    def test_with_no_children_returns_none(self, importer):
        """If there are no child rows, return None"""
        importer.child_works = collate_child_works(
            {
                "ark:/work/1": {
                    "Item ARK": "ark:/work/1",
                    "Parent ARK": "ark:/collection/1",
                    "Thumbnail URL": None,
                    "Title": None,
                    "Object Type": "Work",
                },
            }
        )
        record = {"ark_ssi": "ark:/work/1"}

        result = importer.thumbnail_from_child(record)
        assert result is None


@pytest.fixture
def record():
    return {"iiif_manifest_url_ssi": "http://test.manifest/url/"}


class TestThumbnailFromManifest:
    """Function thumbnail_from_manifest"""

    def test_picks_folio_1r(self, monkeypatch, importer, record):
        "uses the page titled 'f. 001r', if found"
        monkeypatch.setattr(
            feed_ursus.importer.requests, "get", lambda x: fixtures.GOOD_MANIFEST
        )

        result = importer.thumbnail_from_manifest(record)
        assert (
            result
            == "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fzw07hs0c/full/!200,200/0/default.jpg"  # pylint: disable=line-too-long
        )

    def test_picks_first_page(self, monkeypatch, importer, record):
        "uses the first image if 'f. 001r' is not found"
        monkeypatch.setattr(
            feed_ursus.importer.requests,
            "get",
            lambda x: fixtures.MANIFEST_WITHOUT_F001R,
        )

        result = importer.thumbnail_from_manifest(record)
        assert (
            result
            == "https://iiif.sinaimanuscripts.library.ucla.edu/iiif/2/ark%3A%2F21198%2Fz14b44n8%2Fhm957748/full/!200,200/0/default.jpg"  # pylint: disable=line-too-long
        )

    def test_request_fails(self, monkeypatch, importer, record):
        "returns None if HTTP request fails"

        monkeypatch.setattr(
            feed_ursus.importer.requests,
            "get",
            lambda x: fixtures.MockResponse(None, 404),
        )

        result = importer.thumbnail_from_manifest(record)
        assert result is None

    def test_manifest_without_images(self, monkeypatch, importer, record):
        "returns None if manifest contains no images"

        monkeypatch.setattr(
            feed_ursus.importer.requests,
            "get",
            lambda x: fixtures.MANIFEST_WITHOUT_IMAGES,
        )

        result = importer.thumbnail_from_manifest(record)
        assert result is None

    def test_bad_data(self, monkeypatch, importer, record):
        "returns None if manifest isn't parsable"

        monkeypatch.setattr(
            feed_ursus.importer.requests, "get", lambda x: fixtures.BAD_MANIFEST
        )

        result = importer.thumbnail_from_manifest(record)
        assert result is None

    def test_no_manifest_url(self, importer):
        "returns None if the record doesn't include field 'iiif_manifest_url_ssi'"

        result = importer.thumbnail_from_manifest({})
        assert result is None



# mypy: disallow_untyped_defs=False
"""
tests for year_parser.py

Creates a multi-valued 'year_isim' field by parsing 'normalized_date'.
"""

from feed_ursus import year_parser


def test_iso_8601():
    """Parses an iso 8601 standard string"""

    assert year_parser.integer_years(["1941-10-01"]) == [1941]


def test_just_year():
    """Parses a bare year."""

    assert year_parser.integer_years(["1953"]) == [1953]


def test_year_and_month():
    """Parses YYYY-MM"""

    assert year_parser.integer_years(["1953-10"]) == [1953]


def test_multiple_dates():
    """Parses multiple input values into a list of outputs."""

    assert year_parser.integer_years(["1941-10-01", "1935", "1945"]) == [
        1935,
        1941,
        1945,
    ]


def test_empty():
    """Returns an empty list if given an empty input."""
    assert year_parser.integer_years([]) == []


def test_unparseable():
    """Doesn't return anything for unparseable values, but still parses other elements in input."""

    assert year_parser.integer_years(["1953", "[between 1928-1939]"]) == [1953]


def test_range():
    """Parses YYYY/YYYY into a range of years."""

    assert year_parser.integer_years(["1937/1939", "1942/1943"]) == [
        1937,
        1938,
        1939,
        1942,
        1943,
    ]

    # Parses YYY/YYYY into a range of years.

    assert year_parser.integer_years(["990/1000"]) == [
        990,
        991,
        992,
        993,
        994,
        995,
        996,
        997,
        998,
        999,
        1000,
    ]


def test_range_with_months():
    """Months can be included in range elements, but are ingored."""

    assert year_parser.integer_years(["1934-06/1934-07"]) == [1934]


def test_duplicates():
    """Deduplicates elements in output."""

    assert year_parser.integer_years(["1934-06/1935-07", "1934-06-01", "1934"]) == [
        1934,
        1935,
    ]
